from typing import Optional, Tuple, Union, List
from datetime import datetime
import logging
import time
import os

try:
    import slack
    from slack.errors import SlackApiError
except ImportError:
    _has_slack = False
else:
    _has_slack = True

from commands.command import Command
from command_sequence import CommandSequence
from commands.utility_commands import DelayPauseCommand


format = '[%(asctime)s] [%(levelname)-5s]: %(message)s'
log_formatter = logging.Formatter(format)
logging.basicConfig(level=logging.INFO, format=format)


class CommandInvoker:
    """Handles the execution and logging of a command sequence"""
    log_directory = ""

    def __init__(
            self, 
            command_seq: CommandSequence,
            log_to_file: bool = True, 
            log_filename: Optional[str] = None,
            alert_slack: bool = False) -> None:

        if not _has_slack and alert_slack:
            raise ImportError("slackclient module is required to alert slack.")
            
        self._command_seq = command_seq
        self._log_to_file = log_to_file
        self._alert_slack = alert_slack
        self._log_filename = log_filename
        self.log = logging.getLogger(__name__)

        if self._log_to_file:
            if self._log_filename is None:
                timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
                self._log_filename = "logs/" +timestamp

            if not self._log_filename[-4:] == '.log':
                self._log_filename += '.log'
            
            # self._log_filename = self.log_directory + self._log_filename
            
            self._file_handler = logging.FileHandler(self._log_filename)
            self._file_handler.setFormatter(log_formatter)
            if not len(self.log.handlers):
                self.log.addHandler(self._file_handler)

        if self._alert_slack:
            self._slack_token = os.environ.get('SLACK_BOT_TOKEN')
            self._slack_client = slack.WebClient(token=self._slack_token)

    def invoke_commands(self) -> bool:
        """Iterate through the command sequence and execute each command.

        Returns
        -------
        bool
            Whether all the commands executed successfully or not
        """
        # go through each device and make sure uninitialized?
        if self._log_to_file:
            print("")
            print("Log messages will be saved to: " + str(self._log_filename))
            print("")
        else:
            print("")
            print("Log messages will not be saved")
            print("")

        has_error = False

        self.log.info("")
        self.log_command_names(unloop=False)
        self.log.info("="*20 + "BEGINNING OF COMMAND SEQUENCE EXECUTION" + "="*20)

        command_generator = self._command_seq.yield_next_command()

        for command in command_generator:
            # Check the command is a Command otherwise the generator yielded a (False, error message)
            if not isinstance(command, Command):
                error_message = command[1]
                self.log.error("Command sequence error: " + error_message)
                if self._alert_slack:
                    self.log.info("Sending command details to slack.")
                    self.alert_slack_message("Command sequence error: " + error_message)
                    self.upload_log_slack()
                has_error = True
                break

            # Process the command's start delay
            delay = command._params['delay']
            if isinstance(delay, float) or isinstance(delay, int):
                if delay > 0.0:
                    self.log.info("DELAY   -> " + str(delay))
                    time.sleep(delay)
            elif delay == "PAUSE" or delay == "P":
                self.log.info("PAUSE   -> Waiting for user to press enter")
                print('')
                print("Press ENTER to continue, type 'quit' to terminate execution immediately:")
                userinput = input()
                if userinput == "quit":
                    self.log.info("PAUSE   -> User terminated execution early by entering 'quit'")
                    has_error = True # Not really an error but returning False since an early termination (even if intentional) may disrupt a higher workflow
                    break
                else:
                    self.log.info("PAUSE   -> User continued command execution")

            # If it is a delay or pause command skip to next loop to suppress its COMMAND ->, RESULT -> logging
            if isinstance(command, DelayPauseCommand):
                continue

            # Execute the command
            self.log.info("COMMAND -> " + command.name)
            command.execute()

            # Check result
            if command.result.was_successful:
                self.log.info("RESULT  -> " + str(command.result.was_successful) + ", " + command.result.message)
            else:
                self.log.error("RESULT  -> " + str(command.result.was_successful) + ", " + command.result.message)
                self.log.error("Received False result. Terminating command execution early!")
                if self._alert_slack:
                    self.log.info("Sending command details to slack.")
                    self.alert_slack_command(command)
                    self.upload_log_slack()
                has_error = True
                break
            # Go to next command
        # Finished command list execution
        self.log.info("="*20 + "END OF COMMAND SEQUENCE EXECUTION" + "="*20)
        self.log.info("")
        if self._log_to_file:
            print("")
            print("Log messages saved to: " + str(self._log_filename))
            print("")
        else:
            print("")
            print("Log messages were not saved")
            print("")

        if has_error:
            return False
        else:
            return True

    def alert_slack_command(self, command: Command):
        """Attempt to send an error message regarding a command's failure to a designated slack channel.

        Parameters
        ----------
        command : Command
            The command that failed its execution.
        """
        try:
            response = self._slack_client.chat_postMessage(
                channel="printer-bot-test",
                text=("Error in the following command execution:\n" 
                    "COMMAND -> " + command.name + "\n" 
                    "RESULT  -> " + str(command.result.was_successful) + ", " + command.result.message + "\n" 
                    "See log file \"" + str(self._log_filename) + "\" for more details.")
                    )  
        except SlackApiError as inst:
            self.log.error("Could not send message to slack: " + inst.response['error'])

    def alert_slack_message(self, message: str):
        """Attempt to send a message to a designated slack channel.

        Parameters
        ----------
        message : str
            The message to send
        """
        try:
            response = self._slack_client.chat_postMessage(
                channel="printer-bot-test",
                text=message)  
        except SlackApiError as inst:
            self.log.error("Could not send message to slack: " + inst.response['error'])
    
    def upload_log_slack(self):
        """Attempt to upload the invoker's log file to a designated slack channel."""
        try:
            response = self._slack_client.files_upload(    
                file=self._log_filename,
                initial_comment='Uploading log file with error: ' + self._log_filename,
                channels='printer-bot-test'
            )
        except SlackApiError as inst:
            self.log.error("Could not upload log file: " + inst.response['error'])

    def log_command_names(self, unloop: bool = False):
        """Log the names of each command in the sequence

        Parameters
        ----------
        unloop : bool, optional
            Whether to unloop the commands sequence or not, by default False
        """
        self.log.info("="*20 + "LIST OF COMMAND NAMES" + "="*20)
        for name in self._command_seq.get_command_names(unloop):
            self.log.info(name)
        self.log.info("")
        self.log.info("(Number of iterations: " + str(self._command_seq.num_iterations) + ")")
        self.log.info("="*20 + "END OF COMMAND NAMES" + "="*20)

    def log_command_names_descriptions(self, unloop: bool = False):
        """Log the names and descriptions of each command in the sequence

        Parameters
        ----------
        unloop : bool, optional
            Whether to unloop the commands sequence or not, by default False
        """
        self.log.info("="*20 + "LIST OF COMMAND NAMES/DESCRIPTIONS" + "="*20)
        for name_desc in self._command_seq.get_command_names_descriptions(unloop):
            self.log.info(name_desc[0])
            self.log.info(name_desc[1])
        self.log.info("")
        self.log.info("(Number of iterations: " + str(self._command_seq.num_iterations) + ")")
        self.log.info("="*20 + "END OF COMMAND NAMES/DESCRIPTIONS" + "="*20)

    def get_log_messages(self):
        """Get all log messages as a list.

        Returns
        -------
        list
            List of log messages
        """
        log_messages = []
        for handler in self.log.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.flush()  # Ensure all messages are written to the log file
                with open(handler.baseFilename, 'r') as log_file:
                    log_messages.extend(log_file.readlines())
        return log_messages

    def clear_log_file(self):
        """Clear the log file by truncating its content."""
        if self._log_to_file:
            if os.path.exists(self._log_filename):
                with open(self._log_filename, 'w') as log_file:
                    log_file.truncate(0)
                self.log.info("Log file cleared.")
            else:
                self.log.warning("Log file does not exist.")
        else:
            self.log.warning("Log file is not being used.")



# experiment name, id
# is logging, log file, delay between commands, or delay list
# delay list can have a value like -1 or a str to indicate that 
# we wait for user input before proceeding or type quit to terminate early 
# this could also be implemented as a Command that waits for input and returns
#  consider also ctrl c exception termination, safely terminate and log
# invoker observer/inspecter for more complex bookkeeping?

# for params such as experiment name, id, logfile, should these be passed to the invoker constructor OR to the invoke_command method

# for parallel processes with threading consider branches and loop done with the follow command nodes: jump, conditional jump, label, end (ala assembly, exapunks)
# example of conditional jump, prompt user for y/n to jump to a previous label to loop, or jump to a future label to skip some commands
# on loops we can either do the same thing or change the arguments to command (e.g. to explore how printing speed changes each loop). 
# This can be done with a preset that is passed (e.g. a list of parameters corresponding to each loop)
# or this can be done live, e.g. based on equation/condition or based on user input or based on input from a machine learning optimizer
# how to check against infinite loop
# how to check that an execution in a branch doesnt interfere with another branch? the root cannot proceed until its next branch no longer uses receivers that the root will need, and so forth
#  and the root can go as far as it can as long as it uses receivers not used by the next branch?
# how to control threading?, an invoked can call a new thread when it hits a new branch node (or the node is a command itself to start a new thread)
# or an invoker manager or the client sees a new branch node and creates a new invoker and passes it the branch list
# but what if a branch within a branch? need some sort of composite/recursive approach

# how to deal with commands with arguments that change every loop? 
# Should we have command objects that persist throughout loops? if so, should they keep track of their own execution count? (this probably makes writing command classes more complicated)
# should we generate new commands instead, cloning the loop section on each iteration with new commands?
# whose job is it to change the command or generate new commands, the client or invoker
# and how do we deal with commands with arguments that come from user input on each loop?
# how can we make the user input argument and preset list argument methods as similar as possible?
# are commands changable after instantiation or should we just destroy and remake them? 
# if changable then we need update functions/getters to update name/description when arguments change which makes them more complicated
# if we should destroy and remake, we might as well make them when we are ready (after taking user input). 
# But this means the client doesnt have a singular invoker_command method call, and we should always push new commands to the stack/queue and have invoker invoke them
# this means the invoker needs to wait when reaching the end of its current command list, meaning the client needs a definite way to terminate invoking on the invoker



import time
from typing import Optional, Tuple, Union
import functools

from .device import SerialDevice, check_initialized, check_serial


def check_axis_num(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # could be a kwarg or arg
        if 'axis_number' in kwargs:
            axis_number = kwargs['axis_number']
        else:
            axis_number = args[0]
            
        if not self.is_axis_num_valid(axis_number):
            return (False, "Axis number is not valid or not part of passed tuple during construction.")
        return func(self, *args, **kwargs)
    return wrapper

class NewportESP301(SerialDevice):
    def __init__(
            self, 
            name: str,
            port: str,
            baudrate: int = 921600,
            timeout: Optional[float] = 1.0,
            axis_list: Tuple[int, ...] = (1,),
            default_speed: float = 20.0,
            poll_interval: float = 0.1):

        super().__init__(name, port, baudrate, timeout)
        self._axis_list = axis_list
        self._default_speed = default_speed #make list
        # self._default_speed_list = defaults_speed_list
        self._poll_interval = poll_interval
        self._max_speed = 200.0 # make list
        # self._max_speed_list = max_speed_list

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self._name,
            "port": self._port,
            "baudrate": self._baudrate,
            "timeout": self._timeout,
            "axis_list": self._axis_list,
            "default_speed": self._default_speed,
            "poll_interval": self._poll_interval,
        }
        return args_dict

    def update_init_args(self, args_dict: dict):
        self._name = args_dict["name"]
        self._port = args_dict["port"]
        self._baudrate = args_dict["baudrate"]
        self._timeout = args_dict["timeout"]
        self._axis_list = args_dict["axis_list"]
        self._default_speed = args_dict["default_speed"]
        self._poll_interval = args_dict["poll_interval"]

    @property
    def default_speed(self) -> float:
        return self._default_speed

    @default_speed.setter
    def default_speed(self, speed: float):
        if speed > 0.0 and speed < self._max_speed:
            self._default_speed = speed

    # check_error already has serial check
    # easier to just set is_intialized False at the very beginning
    # do for all receivers
    def initialize(self) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")

        was_successful, message = self.check_error() # just used to flush error and serial input buffer if there is an error
        if not was_successful:
            return (was_successful, message)

        self.ser.reset_input_buffer() # flush the serial input buffer even if there was no error

        for axis in self._axis_list:
            # Make sure axis motor is turned on
            was_turned_on, message = self.axis_on(axis)
            if not was_turned_on:
                self._is_initialized = False
                return (was_turned_on, message)
            # set units to mm, homing value to 0, set max speed, set current speed 
            command = str(axis) + "SN2;" + str(axis) + "SH0;" + str(axis) + "VU" + str(self._max_speed) + ";" + str(axis) + "VA" + str(self.default_speed) + "\r"
            self.ser.write(command.encode('ascii'))

        # Make sure initialization of settings was successful
        was_successful, message = self.check_error()
        if not was_successful:
            self._is_initialized = False
            return (was_successful, message)

        for axis in self._axis_list:
            was_homed, message = self.home(axis)
            if not was_homed:
                self._is_initialized = False
                return (was_homed, message)
    
        self._is_initialized = True
        return (True, "Successfully initialized axes by setting units to mm, settings max/current speeds, and homing. Current position set to zero.")

    # move_speed_absolute already has serial check
    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")

        for axis in self._axis_list:
            was_zeroed, message = self.move_speed_absolute(0.0, speed=None, axis_number=axis)
            if not was_zeroed:
                return (was_zeroed, message)

        if reset_init_flag:
            self._is_initialized = False

        return (True, "Successfully deinitialized axes by moving to position zero.")

    # make a home_all function
    @check_serial
    def home(self, axis_number: int) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")

        command = str(axis_number) + "OR4\r"
        self.ser.write(command.encode('ascii'))

        while self.is_any_moving():
            time.sleep(self._poll_interval)
        # pause one more time in case motor stopped moving but position has not been reset yet     
        time.sleep(self._poll_interval)

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)
        else:
            return (True, "Successfully homed axes " + str(axis_number))

    # Consider a decorator for checks?
    @check_serial
    @check_initialized
    @check_axis_num
    def move_speed_absolute(self, axis_number: int = 1, position: Optional[float] = None, speed: Optional[float] = None) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        # #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_axis_num_valid(axis_number):
        #     return (False, "Axis number is not valid or not part of passed tuple during construction.")
        # if not self._is_initialized:
        #     return (False, "ESP301 axes are not initialized.")
        
        # I want axis number to be the first arg so the decorator can pick it up as arg[0]
        # but I also want axis_number to have a default value of 1, so position needs a default value now
        if position is None:
            return (False, "Position was not specified")

        if speed is None:
            speed = self._default_speed

        command = str(axis_number) + "VA" + str(speed) +"\r"
        self.ser.write(command.encode('ascii'))

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)

        if position >= 0.0:
            sign = "+"
        else:
            sign = "-"

        # removed the WS command because it causes timeouts when checking if moving 
        # command = str(axis_number) + "PA" + sign + str(abs(position)) + ";" + str(axis_number) + "WS\r"
        command = str(axis_number) + "PA" + sign + str(abs(position)) + "\r"
        self.ser.write(command.encode('ascii'))

        while self.is_moving(axis_number):
            time.sleep(self._poll_interval)

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)
        else:
            return (True, "Successfully completed absolute move at " + str(position))

    @check_serial
    @check_initialized
    @check_axis_num
    def move_speed_relative(self, axis_number: int = 1, distance: Optional[float] = None, speed: Optional[float] = None) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        # #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_axis_num_valid(axis_number):
        #     return (False, "Axis number is not valid or not part of passed tuple during construction.")
        # if not self._is_initialized:
        #     return (False, "ESP301 axes are not initialized.")
        if distance is None:
            return (False, "Distance was not specified")
        
        if speed is None:
            speed = self._default_speed

        command = str(axis_number) + "VA" + str(speed) +"\r"
        self.ser.write(command.encode('ascii'))

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)

        if distance >= 0.0:
            sign = "+"
        else:
            sign = "-"

        # removed the WS command because it causes timeouts when checking if moving 
        # command = str(axis_number) + "PR" + sign + str(abs(distance)) + ";" + str(axis_number) + "WS\r"
        command = str(axis_number) + "PR" + sign + str(abs(distance)) + "\r"

        self.ser.write(command.encode('ascii'))

        while self.is_moving(axis_number):
            time.sleep(self._poll_interval)

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)
        else:
            return (True, "Successfully completed relative move by " + str(distance))
        

    def is_axis_num_valid(self, axis_number: int) -> bool:
        if axis_number in self._axis_list:
            return True
        else:
            return False
    
    # check axis num
    @check_serial
    @check_axis_num
    def is_moving(self, axis_number: int = 1) -> bool:
        # if not self.ser.is_open:
        #     return False
        # else:
        command = str(axis_number) + "MD?\r"
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()

        if response.strip().decode('ascii') == '0':
            # motion is not done = is moving
            return True
        else:
            # includes timeout case
            return False

    def is_any_moving(self) -> bool:
        is_moving_list = []
        for ndx, axis_number in enumerate(self._axis_list):
            command = str(axis_number) + "MD?\r"
            self.ser.write(command.encode('ascii'))
            response = self.ser.readline()

            if response.strip().decode('ascii') == '0':
                is_moving_list.append(True)
            else:
                is_moving_list.append(False)

        if any(is_moving_list):
            return True
        else: 
            return False

    @check_serial
    def check_error(self) -> Tuple[bool, str]:
        # not needed for queries, but use when instructing to do something
        
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")

        command = "TB?\r"
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()

        if response == b'':
            return (False, "Response timed out.")
        
        response = response.strip().decode('ascii')

        if response[0] == '0':
            return (True, "No errors.")
        else:
            # flush the error buffer
            for n in range(10):
                self.ser.write(command.encode('ascii'))
                self.ser.readline()
            # flush the serial input buffer
            time.sleep(0.1)
            self.ser.reset_input_buffer()
            return (False, response)
    
    @check_serial
    @check_axis_num
    def position(self, axis_number: int = 1) -> Tuple[bool, Union[str, float]]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_axis_num_valid(axis_number):
        #     return (False, "Axis number is not valid or not part of passed tuple during construction.")

        command = str(axis_number) + "TP\r"
        self.ser.write(command.encode('ascii'))
        position_str = self.ser.readline()
        if position_str == b'':
            return (False, "Response timed out.")
        else:    
            return (True, float(position_str.strip().decode('ascii')))

    @check_serial
    @check_axis_num
    def axis_on(self, axis_number: int = 1) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_axis_num_valid(axis_number):
        #     return (False, "Axis number is not valid or not part of passed tuple during construction.")

        command = str(axis_number) + "MO\r"
        self.ser.write(command.encode('ascii'))

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)

        command = str(axis_number) + "MO?\r"
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()

        if response.strip().decode('ascii') == '1':
            return (True, "Axis " + str(axis_number) + " motor successfully turned ON.")
        else:
            # also means timeout
            return (False, "Axis " + str(axis_number) + " motor failed to turned ON.")

    @check_serial
    @check_axis_num
    def axis_off(self, axis_number: int = 1) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_axis_num_valid(axis_number):
        #     return (False, "Axis number is not valid or not part of passed tuple during construction.")

        command = str(axis_number) + "MF\r"
        self.ser.write(command.encode('ascii'))

        was_successful, message = self.check_error()
        if not was_successful:
            return (was_successful, message)

        command = str(axis_number) + "MF?\r"
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()

        if response.strip().decode('ascii') == '0':
            return (True, "Axis " + str(axis_number) + " motor successfully turned OFF.")
        else:
            # also means timeout
            return (False, "Axis " + str(axis_number) + " motor failed to turned OFF.")



# import sys
# import os
from lib2to3.pytree import Base
import time
import threading
from typing import Tuple
import yaml

# from aiohttp import TraceRequestChunkSentParams

from kortex_api.TCPTransport import TCPTransport
from kortex_api.RouterClient import RouterClient, RouterClientSendOptions
from kortex_api.SessionManager import SessionManager
from kortex_api.autogen.messages import Session_pb2

from kortex_api.autogen.client_stubs.BaseClientRpc import BaseClient
from kortex_api.autogen.client_stubs.BaseCyclicClientRpc import BaseCyclicClient
from kortex_api.autogen.messages import Base_pb2 #, BaseCyclic_pb2, Common_pb2


from .device import Device, check_initialized

TCP_PORT = 10000
# UDP_PORT = 10001


class KinovaArm(Device):
    pose_dict_file = 'robot_arm_poses.yaml'
    
    def __init__(
            self, 
            name: str, 
            ip: str = '192.168.1.10',
            username: str = 'admin',
            password: str = 'admin',
            action_timeout: float = 20.0,
            proportional_gain: float = 2.0):
        
        super().__init__(name)
        self._ip = ip
        self._port = TCP_PORT
        self._username = username
        self._password = password
        self._action_timeout = action_timeout
        self._proportional_gain = proportional_gain
        
        with open(self.pose_dict_file) as file:
            # not safe loader
            self._post_dict = yaml.load(file, Loader=yaml.Loader)
        
        self._session_manager = None
        # move to connect?
        self._transport = TCPTransport()
        self._router = RouterClient(self._transport, RouterClient.basicErrorCallback)
        
    def get_init_args(self) -> dict:
        args_dict = {
            "name": self.name,
            "ip": self._ip,
            "username": self._username,
            "password": self._password,
            "action_timeout": self._action_timeout,
            "proportional_gain": self._proportional_gain
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.name = args_dict["name"]
        self._ip = args_dict["ip"]
        self._username = args_dict["username"]
        self._password = args_dict["password"]
        self._action_timeout = args_dict["action_timeout"]
        self._proportional_gain = args_dict["proportional_gain"]

    def connect(self) -> Tuple[bool, str]:
        self._transport.connect(self._ip, self._port)

        self._session_info = Session_pb2.CreateSessionInfo()
        self._session_info.username = self._username
        self._session_info.password = self._password
        self._session_info.session_inactivity_timeout = 10000   # (milliseconds)
        self._session_info.connection_inactivity_timeout = 2000 # (milliseconds)

        self._session_manager = SessionManager(self._router)
        self._session_manager.CreateSession(self._session_info)
        
        self._base = BaseClient(self._router)
        self._base_cyclic = BaseCyclicClient(self._router)
        
        # need to check if connection was successful?
        return (True, "Successfully connected to Kinova robot arm.")
    
    def disconnect(self) -> Tuple[bool, str]:
        if self.session_manager is not None:
            router_options = RouterClientSendOptions()
            router_options.timeout_ms = 1000 
            
            self._session_manager.CloseSession(router_options)

        self._transport.disconnect()
        
        return (True, "Successfully disconnected from Kinova robot arm.")
    
    def initialize(self) -> Tuple[bool, str]:
        was_homed, message = self.home()
        if not was_homed:
            return (False, message)
        
        was_opened, message = self.open_gripper()
        if not was_opened:
            return (False, message)

        self._is_initialized = True
        return (was_homed, "Successfully initialized Kinova arm and gripper.")
    
    def deinitialize(self) -> Tuple[bool, str]:
        pass
        
    
    def home(self) -> Tuple[bool, str]:
        # Make sure the arm is in Single Level Servoing mode
        base_servo_mode = Base_pb2.ServoingModeInformation()
        base_servo_mode.servoing_mode = Base_pb2.SINGLE_LEVEL_SERVOING
        self._base.SetServoingMode(base_servo_mode)
        
        # Move arm to ready position
        # print("Moving the arm to a safe position")
        action_type = Base_pb2.RequestedActionType()
        action_type.action_type = Base_pb2.REACH_JOINT_ANGLES
        action_list = self._base.ReadAllActions(action_type)
        action_handle = None
        
        for action in action_list.action_list:
            # if action.name == "Home":
            if action.name == 'Above Fork Pickup':
                action_handle = action.handle

        if action_handle is None:
            return (False, "Can't reach safe position")

        e = threading.Event()
        notification_handle = self._base.OnNotificationActionTopic(
            self.check_for_end_or_abort(e),
            Base_pb2.NotificationOptions()
        )

        self._base.ExecuteActionFromReference(action_handle)
        finished = e.wait(self._action_timeout)
        self._base.Unsubscribe(notification_handle)

        if finished:
            return (True, "Safe position reached")
        else:
            return (False, "Timeout on action notification wait")

    def execute_action(self, action_name: str):
        # Make sure the arm is in Single Level Servoing mode
        base_servo_mode = Base_pb2.ServoingModeInformation()
        base_servo_mode.servoing_mode = Base_pb2.SINGLE_LEVEL_SERVOING
        self._base.SetServoingMode(base_servo_mode)
        
        # Move arm to ready position
        # print("Moving the arm to a safe position")
        action_type = Base_pb2.RequestedActionType()
        # action_type.action_type = Base_pb2.REACH_JOINT_ANGLES
        action_type.action_type = Base_pb2.REACH_POSE
        action_list = self._base.ReadAllActions(action_type)
        action_handle = None
        
        for action in action_list.action_list:
            if action.name == action_name:
                action_handle = action.handle

        # if not found try to find in reach joint angles
        if action_handle is None:
            action_type.action_type = Base_pb2.REACH_JOINT_ANGLES
        action_list = self._base.ReadAllActions(action_type)
        for action in action_list.action_list:
            if action.name == action_name:
                action_handle = action.handle

        if action_handle is None:
            return (False, "Did not find action of name " + action_name)

        e = threading.Event()
        notification_handle = self._base.OnNotificationActionTopic(
            self.check_for_end_or_abort(e),
            Base_pb2.NotificationOptions()
        )

        self._base.ExecuteActionFromReference(action_handle)
        finished = e.wait(self._action_timeout)
        self._base.Unsubscribe(notification_handle)

        time.sleep(0.2)

        if finished:
            return (True, action_name + " action completed")
        else:
            return (False, "Timeout on action notification wait")

    @check_initialized
    def move_arm_angular(self, pose_name: str):
        action = Base_pb2.Action()
        action.name = "Angular action movement"
        action.application_data = ""

        actuator_count = self._base.GetActuatorCount()

        pose_joint_values = self._post_dict[pose_name]
        
        # Place arm straight up
        for joint_id in range(actuator_count.count):
            joint_angle = action.reach_joint_angles.joint_angles.joint_angles.add()
            joint_angle.joint_identifier = joint_id
            joint_angle.value = pose_joint_values[joint_id]

        e = threading.Event()
        notification_handle = self._base.OnNotificationActionTopic(
            self.check_for_end_or_abort(e),
            Base_pb2.NotificationOptions()
        )
        
        # print("Executing action")
        self._base.ExecuteAction(action)

        # print("Waiting for movement to finish ...")
        finished = e.wait(self._action_timeout)
        self._base.Unsubscribe(notification_handle)

        if finished:
            return (True, "Angular movement completed")
        else:
            return (False, "Timeout on action notification wait")

    # def example_cartesian_action_movement(base, base_cyclic):
        
    #     print("Starting Cartesian action movement ...")
    #     action = Base_pb2.Action()
    #     action.name = "Example Cartesian action movement"
    #     action.application_data = ""

    #     feedback = base_cyclic.RefreshFeedback()

    #     cartesian_pose = action.reach_pose.target_pose
    #     cartesian_pose.x = feedback.base.tool_pose_x          # (meters)
    #     cartesian_pose.y = feedback.base.tool_pose_y - 0.1    # (meters)
    #     cartesian_pose.z = feedback.base.tool_pose_z - 0.2    # (meters)
    #     cartesian_pose.theta_x = feedback.base.tool_pose_theta_x # (degrees)
    #     cartesian_pose.theta_y = feedback.base.tool_pose_theta_y # (degrees)
    #     cartesian_pose.theta_z = feedback.base.tool_pose_theta_z # (degrees)

    #     e = threading.Event()
    #     notification_handle = base.OnNotificationActionTopic(
    #         check_for_end_or_abort(e),
    #         Base_pb2.NotificationOptions()
    #     )

    #     print("Executing action")
    #     base.ExecuteAction(action)

    #     print("Waiting for movement to finish ...")
    #     finished = e.wait(TIMEOUT_DURATION)
    #     base.Unsubscribe(notification_handle)

    #     if finished:
    #         print("Cartesian movement completed")
    #     else:
    #         print("Timeout on action notification wait")
    #     return finished

    # go to position at speed (requires you to know current pos because of speed sign)
    # close until stop at speed (have default)
    def open_gripper(self, grip_speed: float = 0.2):
        grip_speed = abs(grip_speed)

        gripper_command = Base_pb2.GripperCommand()
        finger = gripper_command.gripper.finger.add()
        finger.finger_identifier = 1

        gripper_command.mode = Base_pb2.GRIPPER_SPEED
        finger.value = grip_speed
        self._base.SendGripperCommand(gripper_command)
        gripper_request = Base_pb2.GripperRequest()

        # Wait for reported position to be opened
        gripper_request.mode = Base_pb2.GRIPPER_POSITION
        while True:
            gripper_measure = self._base.GetMeasuredGripperMovement(gripper_request)
            if len (gripper_measure.finger):
                # print("Current position is : {0}".format(gripper_measure.finger[0].value))
                if gripper_measure.finger[0].value < 0.05:
                    break
            else: # Else, no finger present in answer, end loop
                break
        
        wait_time = (abs(grip_speed) / 0.2 * 2 + 0.5)
        time.sleep(wait_time)
        return (True, "Successfully opened gripper")

    def close_gripper(self, grip_speed: float = -0.2):
        grip_speed = -1.0 * abs(grip_speed)

        gripper_command = Base_pb2.GripperCommand()
        finger = gripper_command.gripper.finger.add()
        finger.finger_identifier = 1

        gripper_command.mode = Base_pb2.GRIPPER_SPEED
        finger.value = grip_speed
        self._base.SendGripperCommand(gripper_command)
        gripper_request = Base_pb2.GripperRequest()

        # Wait for reported speed to be 0
        gripper_request.mode = Base_pb2.GRIPPER_SPEED
        while True:
            gripper_measure = self._base.GetMeasuredGripperMovement(gripper_request)
            if len (gripper_measure.finger):
                # print("Current speed is : {0}".format(gripper_measure.finger[0].value))
                if gripper_measure.finger[0].value == 0.0:
                    break
            else: # Else, no finger present in answer, end loop
                break
        wait_time = (abs(grip_speed) / 0.2 * 2 + 0.5)
        time.sleep(wait_time) # without this, a close then open in fast succession causes the close to end early
        return (True, "Successfully closed gripper")

    # def move_gripper(self, position: float, speed: float):
    #     # Create the GripperCommand we will send
    #     gripper_command = Base_pb2.GripperCommand()
    #     finger = gripper_command.gripper.finger.add()

    #     # Close the gripper with position increments
    #     # print("Performing gripper test in position...")
    #     gripper_command.mode = Base_pb2.GRIPPER_POSITION
    #     position = 0.00
    #     finger.finger_identifier = 1
    #     while position < 1.0:
    #         finger.value = position
    #         print("Going to position {:0.2f}...".format(finger.value))
    #         self.base.SendGripperCommand(gripper_command)
    #         position += 0.1
    #         time.sleep(1)

    #     # Set speed to open gripper
    #     print ("Opening gripper using speed command...")
    #     gripper_command.mode = Base_pb2.GRIPPER_SPEED
    #     finger.value = 0.1
    #     self.base.SendGripperCommand(gripper_command)
    #     gripper_request = Base_pb2.GripperRequest()

    #     # Wait for reported position to be opened
    #     gripper_request.mode = Base_pb2.GRIPPER_POSITION
    #     while True:
    #         gripper_measure = self.base.GetMeasuredGripperMovement(gripper_request)
    #         if len (gripper_measure.finger):
    #             print("Current position is : {0}".format(gripper_measure.finger[0].value))
    #             if gripper_measure.finger[0].value < 0.01:
    #                 break
    #         else: # Else, no finger present in answer, end loop
    #             break

    #     # Set speed to close gripper
    #     print ("Closing gripper using speed command...")
    #     gripper_command.mode = Base_pb2.GRIPPER_SPEED
    #     finger.value = -0.1
    #     self.base.SendGripperCommand(gripper_command)

    #     # Wait for reported speed to be 0
    #     gripper_request.mode = Base_pb2.GRIPPER_SPEED
    #     while True:
    #         gripper_measure = self.base.GetMeasuredGripperMovement(gripper_request)
    #         if len (gripper_measure.finger):
    #             print("Current speed is : {0}".format(gripper_measure.finger[0].value))
    #             if gripper_measure.finger[0].value == 0.0:
    #                 break
    #         else: # Else, no finger present in answer, end loop
    #             break


    @staticmethod
    def check_for_end_or_abort(e):
        """Return a closure checking for END or ABORT notifications
        Arguments:
        e -- event to signal when the action is completed
            (will be set when an END or ABORT occurs)
        """
        def check(notification, e = e):
            # print("EVENT : " + \
            #     Base_pb2.ActionEvent.Name(notification.action_event))
            if notification.action_event == Base_pb2.ACTION_END \
            or notification.action_event == Base_pb2.ACTION_ABORT:
                e.set()
        return check


from devices.device import Device
from .command import Command, CommandResult
from devices.mts50_z8 import MTS50_Z8


class MTS50_Z8ParentCommand(Command):
    """Parent class for all MTS50_Z8 commands."""

    receiver_cls = MTS50_Z8

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)


class MTS50_Z8Connect(MTS50_Z8ParentCommand):
    """Open a serial port for the stage."""

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.start_serial())


class MTS50_Z8Initialize(MTS50_Z8ParentCommand):
    """Initialize the stage by homing it."""

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())


class MTS50_Z8Deinitialize(MTS50_Z8ParentCommand):
    """Deinitialize the stage."""

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize())


class MTS50_Z8EnableMotor(MTS50_Z8ParentCommand):
    """Enable the stage motor."""

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params["state"] = True

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_enabled_state(self._params["state"]))


class MTS50_Z8DisableMotor(MTS50_Z8ParentCommand):
    """Disable the stage motor."""

    def __init__(self, receiver: MTS50_Z8, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params["state"] = False

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_enabled_state(self._params["state"]))


class MTS50_Z8MoveAbsolute(MTS50_Z8ParentCommand):
    """Move the stage to an absolute position."""

    def __init__(self, receiver: MTS50_Z8, position: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params["position"] = position

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_absolute(self._params["position"]))


class MTS50_Z8MoveRelative(MTS50_Z8ParentCommand):
    """Move the stage by a relative distance."""

    def __init__(self, receiver: MTS50_Z8, distance: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params["distance"] = distance

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_relative(self._params["distance"]))



from .command import Command, CommandResult, CompositeCommand
from devices.newport_esp301 import NewportESP301
from typing import Optional

# Parent class, subclass from Command ABC
class NewportESP301ParentCommand(Command):
    """Parent class for all NewportESP301 commands."""
    receiver_cls = NewportESP301

    def __init__(self, receiver: NewportESP301, **kwargs):
        super().__init__(receiver, **kwargs)

# Recommended command classes
class NewportESP301Connect(NewportESP301ParentCommand):
    """Open the serial port to the ESP301 controller."""

    def __init__(self, receiver: NewportESP301, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.start_serial())

class NewportESP301Initialize(NewportESP301ParentCommand):
    """Initialize the axes by homing them."""
    
    def __init__(self, receiver: NewportESP301, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class NewportESP301Deinitialize(NewportESP301ParentCommand):
    """Deinitialize the axes by moving them to position zero."""
    
    def __init__(self, receiver: NewportESP301, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

# Device-related commands classes
class NewportESP301MoveSpeedAbsolute(NewportESP301ParentCommand):
    """Move axis to absolute position at specific speed (No speed uses default speed)."""
    
    def __init__(self, receiver: NewportESP301, axis_number: int = 1, position: Optional[float] = None, speed: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['position'] = position
        self._params['speed'] = speed
        self._params['axis_number'] = axis_number
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_speed_absolute(self._params['axis_number'], self._params['position'], self._params['speed']))

class NewportESP301MoveSpeedRelative(NewportESP301ParentCommand):
    """Move axis by relative distance at specific speed (No speed uses default speed)."""
    
    def __init__(self, receiver: NewportESP301, axis_number: int = 1, distance: Optional[float] = None, speed: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['distance'] = distance
        self._params['speed'] = speed
        self._params['axis_number'] = axis_number
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_speed_relative(self._params['axis_number'], self._params['distance'], self._params['speed']))

# Example of command with additional logic to determine the returned tuple of (success/fail: bool, success/fail message: str)
class NewportESP301SetDefaultSpeed(NewportESP301ParentCommand):
    """Set the default speed of the axes."""
    
    def __init__(self, receiver: NewportESP301, speed: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['speed'] = speed

    def execute(self) -> None:
        self._receiver.default_speed = self._params['speed']
        # receiver's default_speed has a setter than checks the value is > 0 and < max_speed before setting
        # therefore, we can check if we actually changed the receiver's default speed, if not, it means the speed was out of range
        if self._receiver.default_speed == self._params['speed']:
            self._result = CommandResult(True, "Default speed successfully set to " + str(self._params['speed']))
        else:
            self._result = CommandResult(False, "Failed to set the default speed to " + str(self._params['speed']) + ". The default speed must be > 0 and < " + str(self._receiver._max_speed))

# Derived commands
class NewportESP301HorzMoveSpeedAbsolute(NewportESP301MoveSpeedAbsolute):
    """desc"""
    
    # position no longer can be None here because we dont need axis num in sig
    def __init__(self, receiver: NewportESP301, position: float, speed: Optional[float] = None, **kwargs):
        super().__init__(receiver, 1, position, speed, **kwargs)

# Just for testing composite commands
class NewportESP301Dance(CompositeCommand):
    def __init__(self, receiver, speed, distance, **kwargs):
        super().__init__(**kwargs)
        self.add_command(NewportESP301MoveSpeedRelative(receiver, -distance, speed))
        self.add_command(NewportESP301MoveSpeedRelative(receiver, distance, speed))


from typing import Tuple, Optional

from .command import Command, CommandResult
from devices.stellarnet_spectrometer import StellarNetSpectrometer

class SpectrometerParentCommand(Command):
    """Parent class for all StellarNet Spectrometer commands."""
    receiver_cls = StellarNetSpectrometer

    def __init__(self, receiver: StellarNetSpectrometer, **kwargs):
        super().__init__(receiver, **kwargs)

class SpectrometerInitialize(SpectrometerParentCommand):
    """Initialize spectrometer by verifying connection and getting each spectrometer object and wavelength array."""

    def __init__(self, receiver: StellarNetSpectrometer, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class SpectrometerDeinitialize(SpectrometerParentCommand):
    """Deinitialize the spectrometer, currently does nothing except optionally change init flag."""

    def __init__(self, receiver: StellarNetSpectrometer, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class SpectrometerUpdateDark(SpectrometerParentCommand):
    """Update the stored dark spectra for all spectrometers."""

    def __init__(
            self, 
            receiver: StellarNetSpectrometer, 
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['integration_times'] = integration_times
        self._params['scans_to_avg'] = scans_to_avg
        self._params['smoothings'] = smoothings

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.update_all_dark_spectra(self._params['integration_times'], self._params['scans_to_avg'], self._params['smoothings']))

class SpectrometerUpdateBlank(SpectrometerParentCommand):
    """Update the stored blank spectra for all spectrometers."""

    def __init__(
            self, 
            receiver: StellarNetSpectrometer, 
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['integration_times'] = integration_times
        self._params['scans_to_avg'] = scans_to_avg
        self._params['smoothings'] = smoothings

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.update_all_blank_spectra(self._params['integration_times'], self._params['scans_to_avg'], self._params['smoothings']))

class SpectrometerGetAbsorbance(SpectrometerParentCommand):
    """Calculate and update the stored absorbance spectra, merge spectra, and optionally save to file. No filename = timestamped filename."""

    def __init__(
            self, 
            receiver: StellarNetSpectrometer,
            save_to_file: bool = True, 
            filename: Optional[str] = None,
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['save_to_file'] = save_to_file
        self._params['filename'] = filename
        self._params['integration_times'] = integration_times
        self._params['scans_to_avg'] = scans_to_avg
        self._params['smoothings'] = smoothings

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.get_all_absorbance(self._params['save_to_file'], self._params['filename'], self._params['integration_times'], self._params['scans_to_avg'], self._params['smoothings']))

class SpectrometerShutterIn(SpectrometerParentCommand):
    pass

class SpectrometerShutterOut(SpectrometerParentCommand):
    pass

class SpectrometerLampOn(SpectrometerParentCommand):
    pass

class SpectrometerLampOff(SpectrometerParentCommand):
    pass


from .command import Command, CommandResult
from devices.kinova_arm import KinovaArm

# Parent class, subclass from Command ABC
class KinovaArmParentCommand(Command):
    """Parent class for all KinovaArm commands."""
    receiver_cls = KinovaArm

    def __init__(self, receiver: KinovaArm, **kwargs):
        super().__init__(receiver, **kwargs)

# Recommended command classes
class KinovaArmConnect(KinovaArmParentCommand):
    """Connect to the Kinova arm (TCP)."""

    def __init__(self, receiver: KinovaArm, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.connect())
        
class KinovaArmInitialize(KinovaArmParentCommand):
    def __init__(self, receiver: KinovaArm, **kwargs):
        super().__init__(receiver, **kwargs)
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class KinovaArmMoveArmAngular(KinovaArmParentCommand):
    def __init__(self, receiver: KinovaArm, pose_name: str, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['pose_name'] = pose_name

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_arm_angular(self._params['pose_name']))

class KinovaArmExecuteAction(KinovaArmParentCommand):
    def __init__(self, receiver: KinovaArm, action_name: str, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['action_name'] = action_name

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.execute_action(self._params['action_name']))

class KinovaArmOpenGripper(KinovaArmParentCommand):
    def __init__(self, receiver: KinovaArm, grip_speed: float = 0.2, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['grip_speed'] = grip_speed

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.open_gripper(self._params['grip_speed']))

class KinovaArmCloseGripper(KinovaArmParentCommand):
    def __init__(self, receiver: KinovaArm, grip_speed: float = -0.2, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['grip_speed'] = grip_speed
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.close_gripper(self._params['grip_speed']))



from datetime import datetime
from typing import Optional, Tuple, List

from PIL import Image
from ximea import *

from .device import Device, check_initialized

# Unsure how cooperative xiapi.Camera is so did not use multiple inheritance
# will need to figure out the method resolution order if using multiple inheritance
class XimeaCamera(Device):
    save_directory = 'data/imaging/'

    def __init__(self, name: str):
        super().__init__(name)
        self.cam = xiapi.Camera()
        # defaults, changeable for each instance
        # could use dictionary but lose IDE hinting
        # could consider a namedtuple to keep descriptive context of each param while still being able to use an iterative
        # for now, individual params
        self._default_imgdataformat = 'XI_RGB24'
        self._default_exposure_time = 50000
        self._default_gain = 0.0
        # default wb coeffs below technically constants, leaving uncaptilaized
        self._default_wb_kr = 1.531
        self._default_wb_kg = 1.0
        self._default_wb_kb = 1.305
        # self.set_default_params() # done in initalize because cam is not yet open here

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self.name,
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.name = args_dict["name"]

    # no setter for imgdataformat at the moment
    @property
    def default_imgdataformat(self) -> str:
        return self._default_imgdataformat
    
    @property
    def default_exposure_time(self) -> int:
        return self._default_exposure_time
    
    @default_exposure_time.setter
    def default_exposure_time(self, exposure_time: int):
        if exposure_time > 0:
            self._default_exposure_time = int(exposure_time)

    @property
    def default_gain(self) -> float:
        return self._default_gain

    @default_gain.setter
    def default_gain(self, gain: float):
        # there is an upper limit for this but not 100% sure what it is
        if gain >= 0.0:
            self._default_gain = gain
    
    # no setter for default wb coeffs
    @property
    def default_wb_kr(self) -> float:
        return self._default_wb_kr

    @property
    def default_wb_kg(self) -> float:
        return self._default_wb_kg

    @property
    def default_wb_kb(self) -> float:
        return self._default_wb_kb

    def set_default_params(self):
        self.cam.set_imgdataformat(self._default_imgdataformat)
        self.cam.set_exposure(self._default_exposure_time)
        self.cam.set_gain(self._default_gain)
        self.cam.set_wb_kr(self._default_wb_kr)
        self.cam.set_wb_kg(self._default_wb_kg)
        self.cam.set_wb_kb(self._default_wb_kb)

    def initialize(self, set_defaults: bool = True) -> Tuple[bool, str]:
        try:
            self.cam.open_device()
            # set defaults if flag is True. This should be True the very first time in order to set the default params, but not enforced
            if set_defaults:
                self.set_default_params()
            self._is_initialized = True
        except xiapi.Xi_error as inst:
            self._is_initialized = False
            return (False, "Failed to connect and initialize: " + str(inst))
 
        if set_defaults:
            return (True, "Successfully initialized camera by opening communications and setting defaults.")
        else:
            return (True, "Successfully intitialized camera by opening communications.")

    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        try:
            self.cam.stop_acquisition()
            self.cam.close_device()
        except xiapi.Xi_error as inst:
            return (False, "Failed to deinitialize the camera: " + str(inst))

        if reset_init_flag:
            self._is_initialized = False

        return (True, "Successfully deinitialized camera, communication closed.")

    @check_initialized
    def get_image(
            self, 
            save_to_file: bool = True, 
            filename: str = None, 
            exposure_time: Optional[int] = None, 
            gain: Optional[float] = None, 
            show_pop_up: bool = False) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "Camera is not initialized")

        if exposure_time is None:
            exposure_time = self._default_exposure_time
        if gain is None:
            gain = self._default_gain

        try:
            self.cam.set_exposure(exposure_time)
            self.cam.set_gain(gain)

            img = xiapi.Image()
            self.cam.start_acquisition()
            # exposure time is in microsec, timeout is in millisec, timeout is set to double the exposure time
            self.cam.get_image(img, timeout=(int(exposure_time / 1000 * 2)))
            self.cam.stop_acquisition()
        except xiapi.Xi_error as inst:
            return (False, "Error while getting image: " + str(inst))

        data = img.get_image_data_numpy(invert_rgb_order=True)
        img = Image.fromarray(data, 'RGB')

        if save_to_file:
            
            if filename is None:
                timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
                filename = timestamp
            fullfilename = self.save_directory + filename
            img.save(fullfilename + '.bmp')

            settings = ["image data format = " + self.cam.get_imgdataformat() + "\n",
                "exposure (us) = " + str(self.cam.get_exposure()) + "\n",
                "gain = " + str(self.cam.get_gain()) + "\n",
                "wb_kr = " + str(self.cam.get_wb_kr()) + "\n",
                "wb_kg = " + str(self.cam.get_wb_kg()) + "\n",
                "wb_kb = " + str(self.cam.get_wb_kb()) + "\n"]
            with open(fullfilename + '.txt', 'w') as file:
                file.writelines(settings) 

            return (True, "Successfully saved image and settings to " + str(fullfilename) + ".bmp")

        if show_pop_up:
            img.show()

        return (True, "Successfully took image but did not save.")       

    @check_initialized
    def update_white_balance(self, exposure_time: int = None, gain: Optional[float] = None) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "Camera is not initialized")
        
        try:
            self.cam.set_manual_wb(1)
        except xiapi.Xi_error as inst:
            return (False, "Failed to set white balance: " + str(inst))

        was_successful, result_message = self.get_image(save_to_file=False, filename=None, exposure_time=exposure_time, gain=gain)

        if not was_successful:
            return (was_successful, result_message)

        return (True, "Successfully updated white balance coefficients with image: wb_kr, wb_kg, wb_kb = " + str(self.get_white_balance_rgb_coeffs()))
 
    @check_initialized # is this necessary
    def set_white_balance_manually(self, wb_kr: Optional[float] = None, wb_kg: Optional[float] = None, wb_kb: Optional[float] = None) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "Camera is not initialized")
        
        if wb_kr is None and wb_kg is None and wb_kb is None:
            return (True, "No white balance coefficients were changed. Coefficients are currently: wb_kr, wb_kg, wb_kb = " + str(self.get_white_balance_rgb_coeffs()))
        try:
            if wb_kr is not None:
                self.cam.set_wb_kr(wb_kr)
            if wb_kg is not None:
                self.cam.set_wb_kg(wb_kg)
            if wb_kb is not None:
                self.cam.set_wb_kb(wb_kb)
        except:
            return (False, "Error in setting white balance coefficients. Coefficients are currently: wb_kr, wb_kg, wb_kb = " + str(self.get_white_balance_rgb_coeffs()))
        
        return (True, "Successfully set white balance coefficients manually: wb_kr, wb_kg, wb_kb = " + str(self.get_white_balance_rgb_coeffs()))
    
    @check_initialized # is this necessary
    def reset_white_balance_rgb_coeffs(self) -> Tuple[bool, str]:
        self.cam.set_wb_kr(self._default_wb_kr)
        self.cam.set_wb_kg(self._default_wb_kg)
        self.cam.set_wb_kb(self._default_wb_kb)
        return (True, "Successfully reset white balance coefficients to defaults. Coefficients are currently: wb_kr, wb_kg, wb_kb = " + str(self.get_white_balance_rgb_coeffs()))

    @check_initialized # is this necessary
    def get_white_balance_rgb_coeffs(self) -> List[float]:
        wb = []
        wb.append(self.cam.get_wb_kr())
        wb.append(self.cam.get_wb_kg())
        wb.append(self.cam.get_wb_kb())
        return wb






from typing import Optional, Tuple
from struct import pack, unpack
import time
from .device import SerialDevice, check_serial, check_initialized


class LinearStage150(SerialDevice):
    def __init__(
        self,
        name: str,
        port: str = "'COM6'",
        baudrate: int = 115200,
        timeout: float | None = 0.1,
        destination: int = 0x50,
        source: int = 0x01,
        channel: int = 1,
    ):
        super().__init__(name, port, baudrate, timeout)
        self._destination = destination
        self._source = source
        self._channel = channel
        # print("dest: "+str(self._destination))

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self._name,
            "port": self._port,
            "baudrate": self._baudrate,
            "timeout": self._timeout,
            "destination": self._destination,
            "source": self._source,
            "channel": self._channel,
        }
        return args_dict

    def update_init_args(self, args_dict: dict):
        self._name = args_dict["name"]
        self._port = args_dict["port"]
        self._baudrate = args_dict["baudrate"]
        self._timeout = args_dict["timeout"]
        self._destination = args_dict["destination"]
        self._source = args_dict["source"]
        self._channel = args_dict["channel"]

    @check_serial
    def initialize(self) -> Tuple[bool, str]:
        self._is_initialized = False

        # lts150: initialize, home it (if needed)

        # Home Stage; MGMSG_MOT_MOVE_HOME
        self.ser.write(pack("<HBBBB", 0x0443, self._channel, 0x00, self._destination, self._source))
        print("Homing stage...")

        # Confirm stage homed before advancing; MGMSG_MOT_MOVE_HOMED
        Rx = ""
        Homed = pack("<H", 0x0444)
        while Rx != Homed:
            Rx = self.ser.read(2)
        print("Stage Homed")
        self.ser.flushInput()
        self.ser.flushOutput()

        self._is_initialized = True
        return (True, "Successfully initialized LTS150 by homing it.")
        # return super().initialize()

    def deinitialize(self) -> Tuple[bool, str]:
        # i dont think this is needed: lts150: deinitialize
        # if reset_init_flag: //used in other devices
        self._is_initialized = False
        return (True, "Successfully deinitialized LTS150.")
        # return super().deinitialize()

    @check_serial
    # @check_initialized
    def get_enabled_state(self) -> bool:
        self._is_enabled = False

        # TODO: lts150 get enabled state, MGMSG_MOD_GET_CHANENABLESTATE
        # self.ser.write(pack('<HBBBB',0x0212,self._channel,0x00,self._destination,self._source))

        return self._is_enabled

    def set_enabled_state(self, state: bool) -> Tuple[bool, str]:
        if state:
            self.ser.write(pack("<HBBBB", 0x0210, 1, 0x01, self._destination, self._source))
        else:
            self.ser.write(pack("<HBBBB", 0x0210, 1, 0x02, 0x50, 0x01))
        time.sleep(0.1)

        self.ser.flushInput()
        self.ser.flushOutput()
        return (True, "Successfully set enable state to " + str(state) + ".")

    @check_serial
    def get_position(self) -> float:
        self._position = 0.0
        Device_Unit_SF = 409600
        # MGMSG_MOT_GET_POSCOUNTER
        self.ser.write(pack("<HBBBB", 0x0411, self._channel, 0x00, self._destination, self._source))

        # Read back position returns by the cube; Rx message MGMSG_MOT_GET_POSCOUNTER
        header, chan_dent, position_dUnits = unpack("<6sHI", self.ser.read(12))
        self._position = position_dUnits / float(Device_Unit_SF)

        return self._position

    @check_serial
    # @check_initialized
    def move_absolute(self, position: float) -> Tuple[bool, str]:
        if position > 150:
            return (False, "Position " + str(position) + " is out of range.")

        Device_Unit_SF = 409600
        dUnitpos = int(Device_Unit_SF * position)
        self.ser.write(
            pack(
                "<HBBBBHI",
                0x0453,
                0x06,
                0x00,
                self._destination | 0x80,
                self._source,
                self._channel,
                dUnitpos,
            )
        )

        # Confirm stage completed move before advancing; MGMSG_MOT_MOVE_COMPLETED
        Rx = ""
        Moved = pack("<H", 0x0464)
        while Rx != Moved:
            Rx = self.ser.read(2)

        print("Move Complete")

        self.ser.flushInput()
        self.ser.flushOutput()

        return (
            True,
            "Successfully moved stage to position " + str(position) + "[units].",
        )

    @check_serial
    # @check_initialized
    def move_relative(self, distance: float) -> Tuple[bool, str]:
        if distance + self.get_position() > 150:
            return (
                False,
                "Position " + str(distance + self.get_position()) + " is out of range.",
            )

        Device_Unit_SF = 409600
        dUnitpos = int(Device_Unit_SF * distance)
        self.ser.write(
            pack(
                "<HBBBBHI",
                0x0448,
                0x06,
                0x00,
                self._destination | 0x80,
                self._source,
                self._channel,
                dUnitpos,
            )
        )

        # Confirm stage completed move before advancing; MGMSG_MOT_MOVE_COMPLETED
        Rx = ""
        Moved = pack("<H", 0x0464)
        while Rx != Moved:
            Rx = self.ser.read(2)

        print("Move Complete")

        self.ser.flushInput()
        self.ser.flushOutput()

        return (
            True,
            "Successfully moved stage by distance " + str(distance) + "[units].",
        )



from typing import List

from .command import Command, CommandResult, CompositeCommand
from devices.dummy_motor import DummyMotor

class DummyMotorParentCommand(Command):
    """Parent class for all DummyMotor commands."""
    receiver_cls = DummyMotor

    def __init__(self, receiver: DummyMotor, **kwargs):
        super().__init__(receiver, **kwargs)

class DummyMotorInitialize(DummyMotorParentCommand):
    """Initialize the motor by homing it."""
    
    def __init__(self, receiver: DummyMotor, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        # original method, tuple unpack into pre-existing attributes
        # self._result._was_successful, self._result._message = self._receiver.initialize()

        # alternate method, tuple unpack positionally into constructor
        self._result = CommandResult(*self._receiver.initialize())

class DummyMotorDeinitialize(DummyMotorParentCommand):
    """Deinitialize the motor by moving to position zero."""
    
    def __init__(self, receiver: DummyMotor, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class DummyMotorSetSpeed(DummyMotorParentCommand):
    """Set the speed of the motor."""

    def __init__(self, receiver: DummyMotor, speed: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['speed'] = speed

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_speed(self._params['speed']))

class DummyMotorMoveAbsolute(DummyMotorParentCommand):
    """Move motor to absolute position."""

    def __init__(self, receiver: DummyMotor, position: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['position'] = position

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_absolute(self._params['position']))

class DummyMotorMoveRelative(DummyMotorParentCommand):
    """Move motor by relative distance."""

    def __init__(self, receiver: DummyMotor, distance: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['distance'] = distance

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_relative(self._params['distance']))


# Composite command that may have one receiver, more than one receiver of the same or different type, or no receiver at all
class DummyMotorMoveSpeedAbsolute(CompositeCommand):
    """Temporarily set the speed and move to position. Speed then reverts to its original value."""

    def __init__(self, receiver: DummyMotor, speed: float, position: float, **kwargs):
        super().__init__(**kwargs)
        original_speed = receiver.motor.speed
        self.add_command(DummyMotorSetSpeed(receiver, speed))
        self.add_command(DummyMotorMoveAbsolute(receiver, position))
        self.add_command(DummyMotorSetSpeed(receiver, original_speed))
        self._params['speed'] = speed
        self._params["position"] = position

class DummyMotorMultiMoveAbsolute(CompositeCommand):
    """Move a list of motors to a list of position at a particular speed"""

    def __init__(self, receiver_list: List[DummyMotor], position_list: List[float], speed: float, **kwargs):
        super().__init__(**kwargs)
        for ndx, receiver in enumerate(receiver_list):
            self.add_command(DummyMotorSetSpeed(receiver, speed))
            self.add_command(DummyMotorMoveAbsolute(receiver, position_list[ndx]))

class DummyMotorMultiInitialize(CompositeCommand):
    """Initialize a list of motors"""

    def __init__(self, receiver_list: List[DummyMotor], **kwargs):
        super().__init__(**kwargs)
        for receiver in receiver_list:
            self.add_command(DummyMotorInitialize(receiver))





import time
from typing import Optional, Tuple, List

from .device import SerialDevice, check_initialized, check_serial


# terminology, valve_num or position; steps_per_sec or step_rate, valve or port
# dead volume flags
# double check flowrates are accurate

class PSD6SyringePump(SerialDevice):
    status_dict = {
        '@': "Pump busy - no error",
        '`': "Pump ready - no error",
        'a': "Initialization error - pump failed to initialize",
        'b': "Invalid command - unrecognized command is used.",
        'c': "Invalid operand - invalid parameter is given with a command.",
        'd': "Invalid command sequence - command communication protocol is incorrect",
        'f': "EEPROM failure - EEPROM is faulty",
        'g': "Syringe not initialized - syringe failed to initialize",
        'i': "Syringe overload - syringe encounters excessive back pressure",
        'j': "Valve overload - valve drive encounters excessive back pressure",
        'k': "Syringe move not allowed - valve is in the bypass or throughput position, syringe move commands are not allowed",
        'o': "Pump busy - command buffer is full"
    }
    
    # volume_factor = {'ul': 1.0, 'ml': 1000.0}
    # time_factor = {'s': 1.0, 'min': 1.0/60.0}
    
    def __init__(
            self,
            name: str,
            port: str,
            baudrate: int = 9600,
            timeout: Optional[float] = 10.0,
            stroke_volume: float = 5000.0, 
            stroke_steps: int = 6000,
            default_flowrate: float = 1000.0,
            # volume_unit: str = 'ul',
            # time_unit: str = 's',
            port_dead_volumes: List[float] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            poll_interval: float = 0.1):

        super().__init__(name, port, baudrate, timeout)
        self._stroke_volume = stroke_volume
        self._stroke_steps = stroke_steps
        self._default_flowrate = default_flowrate
        # self._volume_unit = volume_unit
        # self._time_unit = time_unit
        self._port_dead_volumes = port_dead_volumes
        self._poll_interval = poll_interval
        
        # technically if I send a command that is out of range of these values
        # the pump should be able to give me an error message anyways
        # but if I store a default flowrate in this object I'd rather
        # have it be correct, than let the pump handle it because
        # setting a wrong default flowrate here does not yield an error until
        # later when some other command is sent that uses it
        self._max_steps_per_sec = 10000
        self._min_steps_per_sec = 2
        
        # # this is ok to leave out since anything using a position immediately issues a command
        # self._min_position = 0
        # self._max_position = 6000

    def get_init_args(self) -> dict:
        args_dict = {
            'name': self.name,
            'port': self.port,
            'baudrate': self.baudrate,
            'timeout': self.timeout,
            'stroke_volume': self._stroke_volume,
            'stroke_steps': self._stroke_steps,
            'default_flowrate': self._default_flowrate,
            'port_dead_volumes': self._port_dead_volumes,
            'poll_interval': self._poll_interval
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.name = args_dict['name']
        self.port = args_dict['port']
        self.baudrate = args_dict['baudrate']
        self.timeout = args_dict['timeout']
        self._stroke_volume = args_dict['stroke_volume']
        self._stroke_steps = args_dict['stroke_steps']
        self._default_flowrate = args_dict['default_flowrate']
        self._port_dead_volumes = args_dict['port_dead_volumes']
        self._poll_interval = args_dict['poll_interval']

    @property
    def default_flowrate(self) -> float:
        return self._default_flowrate
    
    @default_flowrate.setter
    def default_flowrate(self, flowrate: float):
        # careful units!
        # steps_per_sec = flowrate / self._stroke_volume * self._stroke_steps
        # steps_per_sec = int(round(steps_per_sec))
        steps_per_sec = self.volume_to_step(flowrate)
        if steps_per_sec >= self._min_steps_per_sec and steps_per_sec <= self._max_steps_per_sec:
            self._default_flowrate = flowrate
            
    @property
    def default_step_rate(self) -> int:
        return self.volume_to_step(self._default_flowrate)
            
    
    # for init and deinit, it needs to set the waste valve!, not just left most or right most!!
    # we need to check, does it move valve first? or syringe first? IT MOVES VALVE FIRST
    # either way, enable hfactor, initialize valve, move to waste, then either run Y/Z or init syringe only
    @check_serial
    def initialize(self) -> Tuple[bool, str]:
        self._is_initialized = False
        
        # sets syringe to 0 (max dispense)
        # sets valve to position 1 (Use command Z to set valve to 6 instead)
        command = '/1ZR\r' # sets output to right side
        self.ser.write(command.encode('ascii'))
        is_ready, message = self.check_error_ready()
        if not is_ready:
            return (False, message)
        self._is_initialized = True
        return (True, "Successfully initialized PSD6 syringe pump.")
    
    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        #move to waste first
        if reset_init_flag:
            self._is_initialized = False
        return (True, "Successfully deinitialized PSD6 syringe pump.")
        
        
    
    def volume_to_step(self, volume: float):
        return int(round(volume / self._stroke_volume * self._stroke_steps))
    
    def step_to_volume(self, step: int):
        return float(step / self._stroke_steps * self._stroke_volume)
    
    # change order, valve num before flowrate
    def move_syringe_absolute_volume(self, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None):
        step = self.volume_to_step(volume)
        if flowrate is None:
            step_rate = None
        else:
            step_rate = self.volume_to_step(flowrate)
        return self.move_syringe_absolute_step(step, valve_num, step_rate)

    # dead volume compensation flag
    def infuse_syringe_volume(self, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None):
        step = self.volume_to_step(volume)
        if flowrate is None:
            step_rate = None
        else:
            step_rate = self.volume_to_step(flowrate)
        return self.infuse_syringe_steps(step, valve_num, step_rate)
    
    def withdraw_syringe_volume(self, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None):
        step = self.volume_to_step(volume)
        if flowrate is None:
            step_rate = None
        else:
            step_rate = self.volume_to_step(flowrate)
        return self.withdraw_syringe_steps(step, valve_num, step_rate)
    
    
    # switch order of step and volume display in message, volume first steps in ()
    @check_serial
    @check_initialized
    def move_syringe_absolute_step(self, step: int, valve_num: Optional[int] = None, step_rate: Optional[int] = None):
        if valve_num is not None:
            self.move_valve_position(valve_num)
        if step_rate is None:
            step_rate = self.default_step_rate
        # this is here in case a user manually passes a step_rate
        if isinstance(step_rate, float):
            step_rate = int(round(step_rate))
        
        command = '/1V' + str(step_rate) + 'A' + str(step) + 'R\r'
        self.ser.write(command.encode('ascii'))
        is_ready, message = self.check_error_ready()
        if not is_ready:
            return (False, message)

        volume = self.step_to_volume(step)
        return (True, "Successfully moved syringe to step " + str(step) + ", volume " + str(volume) + " uL.")
       
    @check_serial
    @check_initialized
    def infuse_syringe_steps(self, steps: int, valve_num: Optional[int] = None, step_rate: Optional[int] = None):
        if valve_num is not None:
            self.move_valve_position(valve_num)
        if step_rate is None:
            step_rate = self.default_step_rate
        if isinstance(step_rate, float):
            step_rate = int(round(step_rate))
        
        command = '/1V' + str(step_rate) + 'D' + str(steps) + 'R\r'
        self.ser.write(command.encode('ascii'))
        is_ready, message = self.check_error_ready()
        if not is_ready:
            return (False, message)

        volume = self.step_to_volume(steps)
        return (True, "Successfully infused syringe by steps " + str(steps) + ", volume " + str(volume) + " uL.")
    
    @check_serial
    @check_initialized
    def withdraw_syringe_steps(self, steps: int, valve_num: Optional[int] = None, step_rate: Optional[int] = None):
        if valve_num is not None:
            self.move_valve_position(valve_num)
        if step_rate is None:
            step_rate = self.default_step_rate
        if isinstance(step_rate, float):
            step_rate = int(round(step_rate))
        
        command = '/1V' + str(step_rate) + 'P' + str(steps) + 'R\r'
        self.ser.write(command.encode('ascii'))
        is_ready, message = self.check_error_ready()
        if not is_ready:
            return (False, message)

        volume = self.step_to_volume(steps)
        return (True, "Successfully withdrew syringe by steps " + str(steps) + ", volume " + str(volume) + " uL.")
    
    @check_serial
    @check_initialized
    def move_valve_position(self, valve_num: int):
        command = '/1I' + str(valve_num) + "R\r"
        self.ser.write(command.encode('ascii'))
        is_ready, message = self.check_error_ready()
        if not is_ready:
            return (False, message)
        
        return (True, "Successfully moved valve to position " + str(valve_num))
    
    
    # we need to think how numerical information is returned
    # either return value, or return tuple of (bool, message=value)
    # or consider adding a new attribute to the CommandResult class
    def syringe_position(self):
        pass
    
    def valve_position(self):
        command = '/1?24000R\r'
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()
        position = response.decode('ascii')[3]
        return position

    def port_dead_volume(self, port_num: int):
        # no real need to check port, any command sent regarding a wrong port will be caught by the pump
        # if port_num < 1 or port_num > 6:
        #     pass 
        return self._port_dead_volumes[port_num-1]
    
    def check_error_ready(self, read_first: bool = True) -> Tuple[bool, str]:
        # Read first if checking right after sending a command
        # there should be a response in the input buffer
        if read_first:
            response = self.ser.readline()
            status_byte = response.decode('ascii')[2]
            if status_byte not in '@`':
                # there was a problem with the initial response
                return (False, self.status_dict[status_byte])
        
        # keep querying status until ready or error
        status_byte = '@'    
        while status_byte == '@':
            time.sleep(self._poll_interval)
            status_byte = self.query_status_byte()
            
        message = self.status_dict[status_byte]
        
        if status_byte == '`':
            return (True, message)
        else:
            return (False, message)

    def query_status_byte(self):
        command = "/1QR\r"
        self.ser.write(command.encode('ascii'))
        response = self.ser.readline()
        status_byte = response.decode('ascii')[2]
        return status_byte


# modules for device as of commit 4123ed0

"""Requires 'StandardFirmata' basic example uploaded on Arduino Uno"""
from typing import Tuple, Optional
import time
import pyfirmata

from .device import ArduinoSerialDevice, check_initialized, check_serial


class FestoSolenoidValve(ArduinoSerialDevice):
    def __init__(
        self,
        name: str,
        numchannel: int,
        port: str = "COM5",
        baudrate: int = 9600,
        timeout: float = 0.1,
    ):
        super().__init__(name, port, baudrate, timeout)
        self.board = pyfirmata.Arduino(self.port)
        self.numchannel = numchannel
        self.pin = self.board.get_pin(f"d:{numchannel}:o")

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self.name,
            "numchannel": self.numchannel,
            "port": self.port,
            "baudrate": self.baudrate,
            "timeout": self.timeout,
        }
        return args_dict

    def update_init_args(self, args_dict: dict):
        self.name = args_dict["name"]
        self.numchannel = args_dict["numchannel"]
        self.port = args_dict["port"]
        self.pin = self.board.get_pin(f"d:{self.port}:o")  # TODO: Check if this is necessary
        self.baudrate = args_dict["baudrate"]
        self.timeout = args_dict["timeout"]

    def initialize(self) -> Tuple[bool, str]:
        self._is_initialized = True
        # TODO: solenoid valve initialize
        return (True, "Solenoid valve initialized")

    def deinitialize(self) -> Tuple[bool, str]:
        self._is_initialized = False
        self.board.exit()
        return (True, "Solenoid valve deinitialized")

    @check_serial
    @check_initialized
    def valve_open(self) -> Tuple[bool, str]:
        self.pin.write(1)
        return (True, "Solenoid valve is open")

    @check_serial
    def valve_closed(self) -> Tuple[bool, str]:
        self.pin.write(0)
        return (True, "Solenoid valve is closed")

    @check_serial
    def open_timed(self, time: int) -> Tuple[bool, str]:
        self.pin.write(1)
        self.board.pass_time(time)
        self.pin.write(0)
        return (True, f"Solenoid valve was opened for {time} seconds")



from command_sequence import CommandSequence
from command_invoker import CommandInvoker
import json
from devices.device import SerialDevice
import util
from mongodb_helper import MongoDBHelper
import dash
from dash import dcc
from dash import html
from dash import dash_table
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc
import os, signal
import inspect
from bson.objectid import ObjectId
from console_interceptor import ConsoleInterceptor
from gridfs import GridFS
import base64
import io

try:
    import serial.tools.list_ports
except ImportError:
    _has_serial = False
else:
    _has_serial = True
import typing

if os.path.isfile("pw.txt"):
    with open("pw.txt", "r") as f:
        mongo_username, mongo_password = f.read().split("\n")
else:
    mongo_username = input("Enter MongoDB username: ")
    mongo_password = input("Enter MongoDB password: ")
    with open("pw.txt", "w") as f:
        f.write(mongo_username + "\n" + mongo_password)

print("\nreset complete")
com = CommandSequence()
invoker = CommandInvoker(com, log_to_file=True, log_filename="mylog.log")
invoker.clear_log_file()
invoker.invoking = False
if os.path.isfile("blank.yaml"):
    com.load_from_yaml("blank.yaml")
else:
    with open("blank.yaml", "w") as f:
        f.write(
            """- []
- []
- ALL"""
        )


mongo = MongoDBHelper(
    "mongodb+srv://"
    + mongo_username
    + ":"
    + mongo_password
    + "@diaogroup.nrcgqsq.mongodb.net/?retryWrites=true&w=majority",
    "diaogroup",
)
mongo_gridfs = GridFS(mongo.db, collection="recipes")

app = dash.Dash(
    __name__,
    external_stylesheets=[dbc.themes.BOOTSTRAP],
    use_pages=True,
    prevent_initial_callbacks="initial_duplicate",
)
server = app.server

navbar = dbc.NavbarSimple(
    children=[
        dbc.NavItem(dbc.NavLink("Load", href="/load-recipe", external_link=True)),
        dbc.NavItem(dbc.NavLink("View", href="/view-recipe", external_link=True)),
        dbc.NavItem(dbc.NavLink("Execute", href="/execute-recipe", external_link=True)),
        dbc.NavItem(
            dbc.NavLink("Manual Control", href="/manual-control", external_link=True)
        ),
        dbc.NavItem(
            dbc.NavLink("Database Browser", href="/database", external_link=True)
        ),
        dbc.DropdownMenu(
            children=[
                # dbc.DropdownMenuItem(
                #     "Load Recipe", href="/load-recipe", external_link=True
                # ),
                dbc.DropdownMenuItem(
                    "Real Time Telemetry",
                    href="/real-time-telemetry",
                    external_link=True,
                ),
                dbc.DropdownMenuItem(
                    "Edit Code", href="/edit-recipe", external_link=True
                ),
                dbc.DropdownMenuItem("Document", href="/data", external_link=True),
            ],
            nav=True,
            in_navbar=True,
            label="Tools",
        ),
    ],
    brand="AAMP",
    brand_href="/",
    color="primary",
    dark=True,
    id="navbar",
    className="mb-3",
)


app.layout = html.Div([dcc.Location(id="url"), navbar, dash.page_container])


@app.callback(Output("navbar", "brand"), Input("url", "pathname"))
def print_pagename(url):  # all pages
    print("loading: " + url + "\n")
    return "AAMP"


def update_upstream_recipe_dict():
    print("update_upstream_recipe_dict")
    if "document" in list(com.__dict__.keys()):
        recipe_dict = com.get_recipe()
        com.document["recipe_dict"] = {
            "devices": recipe_dict[0],
            "commands": recipe_dict[1],
            "execution_options": recipe_dict[2],
        }
        mongo.db["recipes"].update_one(
            {"_id": com.document["_id"]}, {"$set": com.document}
        )
        print("successfully updated recipe_dict upstream")
        return True
    else:
        print("com.document not found")
        return False


def update_execution_upstream(execution):
    if "document" in list(com.__dict__.keys()):
        com.document["executions"].append(execution)
        mongo.db["recipes"].update_one(
            {"_id": com.document["_id"]}, {"$set": com.document}
        )
        print("successfully updated execution upstream")
        return True
    else:
        print("com.document not found")
        return False


# ---------------------------------------------------
# Home Page
# ---------------------------------------------------


@app.callback(
    Output("home-recipes-list-table", "data"),
    [Input("home-refresh-list-button", "n_clicks")],
    # prevent_initial_call=True,
)
def fetch_recipe_list(n_clicks):  # homepage
    docs = mongo.find_documents("recipes", {})
    docs = mongo.db["recipes"].find(
        {},
        {
            "_id": 0,
            "file_name": 1,
            "posix_friendly": 1,
            "dash_friendly": 1,
            "python_code": 1,
        },
    )
    print("fetch_recipe_list")
    data = []
    for doc in docs:
        file_name = doc.get("file_name", "")
        posix_friendly = doc.get("posix_friendly", True)
        dash_friendly = doc.get("dash_friendly", False)
        if "python_code" in doc:
            python_code = True
        else:
            python_code = False
        data.append(
            {
                "file_name": file_name,
                "posix_friendly": posix_friendly,
                "dash_friendly": dash_friendly,
                "python_code": python_code,
            }
        )
    return data


@app.callback(
    Output("filename-input", "value"),
    Input("home-recipes-list-table", "active_cell"),
    [State("url", "pathname"), State("home-recipes-list-table", "data")],
    # prevent_initial_call=True,
)
def fill_filename_input(active_cell, url, data):  # homepage
    if str(url) == "/load-recipe":
        if active_cell is not None:
            print("fill_filename_input")
            return data[active_cell["row"]]["file_name"]
        if "document" in list(com.__dict__.keys()):
            print("fill_filename_input")
            return com.document["file_name"]
        return ""


@app.callback(
    [
        Output("home-load-file-alert", "is_open"),
        Output("home-load-file-alert", "children"),
        Output("home-load-file-alert", "color"),
        Output("home-load-file-alert", "duration"),
    ],
    [Input("filename-input-button", "n_clicks")],
    [State("filename-input", "value")],
    prevent_initial_call=True,
)
def get_document_from_db(n_clicks, filename):  # homepage
    print("get_document_from_db")
    if filename is not None and filename != "":
        # Extract the YAML content from the document
        document = mongo.find_documents("recipes", {"file_name": filename})[0]
        invoker.clear_log_file()
        try:
            com.clear_recipe()
            com.load_from_dict(document["recipe_dict"])
            com.document = document
            print("loaded from dict")
            return [True, "Recipe loaded", "success", 10000]

        except Exception as e:
            print("Failed load from dict: " + str(e))
        if os.name == "posix":
            if "posix_friendly" in document and not document["posix_friendly"]:
                return [
                    True,
                    "This recipe is not compatible with your system (POSIX compatability error)",
                    "danger",
                    8000,
                ]
        if (
            document.get("dash_friendly", "") == False
            or document.get("python_code", "") == ""
        ):
            yaml_content = document.get("yaml_data", "")
            # Update the YAML output
            with open("to_load.yaml", "w") as file:
                file.write(yaml_content)
            com.load_from_yaml("to_load.yaml")
            com.document = document
            return [True, "Recipe loaded", "success", 10000]
        else:
            exec(document.get("python_code", ""))
            com.load_from_yaml("to_save.yaml")
            com.document = document

        # com.python_code = document.get("python_code", "")

        return [True, "Recipe loaded", "success", 10000]

    return [True, "No recipe selected", "warning", 3000]


@app.callback(
    [
        Output("home-load-file-alert", "is_open", allow_duplicate=True),
        Output("home-load-file-alert", "children", allow_duplicate=True),
        Output("home-load-file-alert", "color", allow_duplicate=True),
        Output("home-load-file-alert", "duration", allow_duplicate=True),
    ],
    Input("home-create-new-recipe-button", "n_clicks"),
    [State("url", "pathname"), State("filename-input", "value")],
    prevent_initial_call=True,
)
def create_new_recipe_doc(n, url, name):
    if str(url) == "/load-recipe":
        print("create_new_recipe_doc")
        mongo.db["recipes"].insert_one(
            {
                "file_name": name,
                "recipe_dict": {
                    "devices": [],
                    "commands": [],
                    "execution_options": {
                        "output_files": [],
                        "default_execution_record_name": "Execution - " + str(name),
                    },
                },
                "dash_friendly": True,
                "executions": [],
            }
        )

        return [True, "Recipe created", "success", 10000]


# ---------------------------------------------------
# View Recipe Page
# ---------------------------------------------------


@app.callback(
    Output("devices-table-div", "children"),
    [Input("refresh-button1", "n_clicks"), Input("devices-table", "data")],
    [State("devices-table-div", "children")],
)
def update_device_table(n_clicks, data, table):  # view-recipe page
    print("update_device_table")
    # table_data1 = dl5
    table_data1 = com.get_clean_device_list().copy()
    # print(com.device_list[1].get_init_args())
    table_data1_new = []
    for index, list in enumerate(table_data1):
        # table_data1[index][1].update({"device_type": table_data1[index][0]})
        # del table_data1[index][0]
        table_data1_new.append(
            {
                "index": index,
                "device_type": table_data1[index][0],
                "params": str(table_data1[index][1]),
            }
        )

    table_data1 = table_data1_new

    table = dash_table.DataTable(
        id="devices-table",
        data=table_data1,
        columns=[
            {"name": "Index", "id": "index"},
            {"name": "Type", "id": "device_type"},
            {"name": "Parameters", "id": "params"},
        ],
        style_cell={
            "overflow": "hidden",
            "textOverflow": "ellipsis",
            "maxWidth": 0,
            "textAlign": "left",
            "padding": "5px",
        },
        style_cell_conditional=[
            {"if": {"column_id": "index"}, "width": "5%"},
            {"if": {"column_id": "device_type"}, "width": "20%"},
            {"if": {"column_id": "params"}, "width": "70%"},
        ],
        # tooltip_data=[
        #     {
        #         column: {"value": str(value), "type": "markdown"}
        #         for column, value in row.items()
        #     }
        #     for row in table_data1
        # ],
        # tooltip_duration=None,
        # editable = True,
    )
    return table


@app.callback(
    Output("commands-table", "data"),
    Input("save-command-editor", "n_clicks"),
    [
        State("commands-table", "active_cell"),
        State("commands-table", "data"),
        State("command-json-editor", "value"),
    ],
    prevent_initial_call=True,
)
def save_command(n_clicks, active_cell, data, value):  # view-recipe page
    print("save_command")
    if active_cell is not None and data[active_cell["row"]]["params"] != str(
        json.loads(value)
    ):
        com.command_list[data[active_cell["row"]]["index"]][0]._params = eval(value)
        update_upstream_recipe_dict()
        return None
    return data


@app.callback(
    [
        Output("devices-table", "data"),
        Output("view-recipe-alert", "is_open", allow_duplicate=True),
        Output("view-recipe-alert", "children", allow_duplicate=True),
        Output("view-recipe-alert", "color", allow_duplicate=True),
        Output("view-recipe-alert", "duration", allow_duplicate=True),
    ],
    Input("save-device-editor", "n_clicks"),
    [
        State("devices-table", "active_cell"),
        State("devices-table", "data"),
        State("device-json-editor", "value"),
    ],
    prevent_initial_call=True,
)
def save_device(n_clicks, active_cell, data, value):  # view-recipe page
    print("save_device")
    if active_cell is not None and data[active_cell["row"]]["params"] != str(
        json.loads(value)
    ):
        # data_row = data[active_cell["row"]]
        params = eval(value)
        com.device_by_name[params["name"]].update_init_args(params)
        update_success = update_upstream_recipe_dict()
        return None, True, "Device updated.", "success", 3000
    return data, False, "Something went wrong. Device not updated.", "danger", 3000


@app.callback(
    Output("command-json-editor", "value"),
    [Input("command-editor-modal", "is_open")],
    [State("commands-table", "active_cell"), State("commands-table", "data")],
    prevent_initial_call=True,
)
def fill_command_json_editor(is_open, active_cell, data):  # view-recipe page
    if active_cell is not None and is_open:
        print("fill_command_json_editor")
        return json.dumps(eval(data[active_cell["row"]]["params"]), indent=4)

    return ""


@app.callback(
    Output("add-device-dropdown", "options"),
    [Input("device-add-modal", "is_open")],
    [State("devices-table", "active_cell"), State("devices-table", "data")],
    prevent_initial_call=True,
)
def fill_device_add_modal(is_open, active_cell, data):  # view-recipe page
    print("fill_device_add_modal")
    return list(util.devices_ref_redundancy.keys())


@app.callback(
    [Output("add-device-json-editor", "value")],
    [Input("add-device-dropdown", "value"), Input("device-add-modal", "is_open")],
    prevent_initial_call=True,
)
def fill_device_add_json_editor(value, is_open):  # view-recipe page
    print("fill_device_add_json_editor")
    if not is_open or value is None:
        return [""]
    # args_list = inspect.getfullargspec(util.devices_ref_redundancy[value]['obj'].__init__).args
    args_dict = {}
    # for arg in args_list:
    #     if arg != "self" and arg != "name":
    #         args_dict[arg] = None
    #     if arg == "name":
    #         args_dict[arg] = value
    args_list = list(util.devices_ref_redundancy[value]["init"]["args"].keys())
    for arg in args_list:
        args_dict[arg] = util.devices_ref_redundancy[value]["init"]["args"][arg][
            "default"
        ]
    return [(json.dumps(args_dict, indent=4))]


@app.callback(
    [
        Output("device-json-editor", "value"),
        Output("edit-device-serial-ports-info", "children"),
    ],
    [Input("device-editor-modal", "is_open")],
    [State("devices-table", "active_cell"), State("devices-table", "data")],
    prevent_initial_call=True,
)
def fill_device_json_editor(is_open, active_cell, data):  # view-recipe page
    print("fill_device_json_editor")
    if active_cell is not None and is_open:
        if _has_serial and isinstance(
            com.device_by_name[eval(data[active_cell["row"]]["params"])["name"]],
            SerialDevice,
        ):
            ports = serial.tools.list_ports.comports()
            str_ports = ""
            for port, desc, hwid in sorted(ports):
                str_ports += f"{port}: {desc} [{hwid}]\n"
            lines = str_ports.splitlines()
            device_port_html = [
                html.Div(["COM Port Info:"], style={"fontWeight": "bold"})
            ]
            device_port_html.append(html.Div([html.Div(line) for line in lines]))
        else:
            device_port_html = ""
        return (
            json.dumps(eval(data[active_cell["row"]]["params"]), indent=4),
            device_port_html,
        )
    return "", ""


@app.callback(
    [
        Output("save-command-editor", "disabled"),
        Output("edit-command-error", "children"),
    ],
    Input("command-json-editor", "value"),
    State("command-editor-modal", "is_open"),
    prevent_initial_call=True,
)
def enable_save_command_button(value, is_open):  # view-recipe page
    if not is_open:
        return False, ""
    try:
        parsed_json = json.loads(value)
        if parsed_json["delay"] < 0:
            return True, "Delay must be greater than or equal to 0"
        print("enable_save_command_button")
        return False, ""
    except Exception as e:
        if type(e) == json.decoder.JSONDecodeError:
            return True, "Invalid JSON"
        return True, str(type(e))


@app.callback(
    [Output("save-device-editor", "disabled"), Output("edit-device-error", "children")],
    Input("device-json-editor", "value"),
    State("device-editor-modal", "is_open"),
    prevent_initial_call=True,
)
def enable_save_device_button(value, is_open):  # view-recipe page
    if not is_open:
        return False, ""
    try:
        parsed_json = json.loads(value)
        print("enable_save_device_button")
        return False, ""
    except Exception as e:
        if type(e) == json.decoder.JSONDecodeError:
            return True, "Invalid JSON"
        return True, str(type(e))


@app.callback(
    [Output("add-device-editor", "disabled"), Output("add-device-error", "children")],
    [Input("add-device-json-editor", "value"), Input("add-device-dropdown", "value")],
    State("device-add-modal", "is_open"),
    prevent_initial_call=True,
)
def enable_add_device_button(value, device_type, is_open):  # view-recipe page
    if value == "":
        return True, "No device selected"
    if not is_open:
        return False, ""
    try:
        sig = inspect.signature(util.named_devices[device_type].__init__)
        args = {}
        for param in sig.parameters.values():
            arg_type = param.annotation
            args[param.name] = (
                typing.get_args(arg_type)[0]
                if typing.get_origin(arg_type) is typing.Union
                else arg_type
            )
        parsed_json = json.loads(value)
        for key in parsed_json:
            # print("\n" + key)
            # print(
            #     "input: "
            #     + str(type((parsed_json[key])))
            #     + ", expected: "
            #     + str(args[key])
            # )
            if type((parsed_json[key])) != args[key]:
                return False, f"Invalid type for {key}. Expected {str(args[key])}"
            # if not isinstance(parsed_json[key], args[key]):
            #     return True, f"Invalid type for {key}. Expected {str(args[key])}"
        print("enable_add_device_button")
        return False, ""
    except Exception as e:
        if type(e) == json.decoder.JSONDecodeError:
            return True, "Invalid JSON"
        return False, str(type(e))


@app.callback(
    [
        Output("view-recipe-alert", "is_open"),
        Output("view-recipe-alert", "children"),
        Output("view-recipe-alert", "color"),
        Output("view-recipe-alert", "duration"),
    ],
    [Input("add-device-editor", "n_clicks")],
    [
        State("add-device-dropdown", "value"),
        State("url", "pathname"),
        State("add-device-json-editor", "value"),
    ],
    prevent_initial_call=True,
)
def view_recipe_add_device(n, device_type, url, device_dict):
    if str(url) == "/view-recipe":
        com.add_device_from_dict(device_type, json.loads(device_dict))
        update_success = update_upstream_recipe_dict()
        return (
            True,
            f"Added {device_type}. Database updated: {update_success}",
            "success",
            3000,
        )


@app.callback(
    Output("edit-command-button", "disabled"),
    Input("commands-table", "active_cell"),
)
def edit_command_button(table_div_children):  # view-recipe page
    active_cell = table_div_children
    if active_cell is not None:
        print("edit_command_button")
        return False
    else:
        return True


@app.callback(
    Output("edit-device-button", "disabled"),
    Input("devices-table", "active_cell"),
)
def edit_device_button(table_div_children):  # view-recipe page
    active_cell = table_div_children
    if active_cell is not None:
        return False
    else:
        return True


@app.callback(
    Output("delete-device-button", "disabled"),
    Input("devices-table", "active_cell"),
)
def view_recipe_enable_delete_device_button(table_div_children):
    active_cell = table_div_children
    if active_cell is not None:
        return False
    else:
        return True


@app.callback(
    Output("delete-command-button", "disabled"), Input("commands-table", "active_cell")
)
def view_recipe_enable_delete_command_button(table_div_children):
    active_cell = table_div_children
    if active_cell is not None:
        return False
    else:
        return True


@app.callback(
    Output("devices-table", "data", allow_duplicate=True),
    Input("delete-device-button", "n_clicks"),
    [
        State("url", "pathname"),
        State("devices-table", "active_cell"),
        State("devices-table", "data"),
    ],
    prevent_initial_call=True,
)
def view_recipe_delete_device(n, url, active_cell, data):
    if str(url) == "/view-recipe":
        com.remove_device_by_index(data[active_cell["row"]]["index"])
        update_success = update_upstream_recipe_dict()
        return None


@app.callback(
    Output("commands-table", "data", allow_duplicate=True),
    Input("delete-command-button", "n_clicks"),
    [
        State("url", "pathname"),
        State("commands-table", "active_cell"),
        State("commands-table", "data"),
    ],
    prevent_initial_call=True,
)
def view_recipe_delete_command(n, url, active_cell, data):
    if str(url) == "/view-recipe":
        com.remove_command(data[active_cell["row"]]["index"])
        update_success = update_upstream_recipe_dict()
        return None


@app.callback(
    Output("commands-table-div", "children"),
    [Input("refresh-button2", "n_clicks"), Input("commands-table", "data")],
    [State("commands-table-div", "children")],
)
def update_commands_table(n_clicks, data, table):  # view-recipe page
    print("update_commands_table")
    command_list = com.command_list.copy()
    command_params = []
    for index, command in enumerate(command_list):
        temp_dict_command_params = {"command": type(command[0]).__name__}
        temp_dict_command_params.update(
            {"params": str(command[0].get_init_args()), "index": index}
        )
        command_params.append((temp_dict_command_params))
        # else:
        #     command_params.append(command._params)

    table_data2 = command_params
    # add_command_accordian(0, to_add=[dbc.AccordionItem("new new", title="new new", item_id="new new")])

    table = dash_table.DataTable(
        id="commands-table",
        data=table_data2,
        columns=[
            {"name": "Index", "id": "index"},
            {"name": "Command", "id": "command"},
            {"name": "Parameters", "id": "params"},
        ],
        style_cell={
            "overflow": "hidden",
            "textOverflow": "ellipsis",
            "maxWidth": 0,
            "textAlign": "left",
            "padding": "5px",
        },
        style_cell_conditional=[
            {"if": {"column_id": "index"}, "width": "5%"},
            {"if": {"column_id": "command"}, "width": "20%"},
            {"if": {"column_id": "params"}, "width": "70%"},
        ],
        # tooltip_data=[
        #     {
        #         column: {"value": str(value), "type": "markdown"}
        #         for column, value in row.items()
        #     }
        #     for row in table_data2
        # ],
        # tooltip_duration=None,
        # editable = True,
    )
    return table


@app.callback(
    Output("view-recipe-command-add-modal", "is_open"),
    Input("add-command-open-modal-button", "n_clicks"),
    [State("url", "pathname")],
    prevent_initial_call=True,
)
def view_recipe_open_add_command_modal(n, url):
    if url == "/view-recipe":
        return True


@app.callback(
    Output("view-recipe-add-command-device-dropdown", "options"),
    Input("view-recipe-command-add-modal", "is_open"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def view_recipe_fill_add_command_device_dropdown(is_open, url):
    if url == "/view-recipe":
        print("view_recipe_fill_add_command_device_dropdown")
        return list(util.devices_ref_redundancy.keys())


@app.callback(
    Output("view-recipe-add-command-command-dropdown", "options"),
    Input("view-recipe-add-command-device-dropdown", "value"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def view_recipe_fill_add_command_command_dropdown(device_type, url):
    if url == "/view-recipe":
        if device_type is None or device_type == "":
            return []
        return list(util.devices_ref_redundancy[device_type]["commands"].keys())


@app.callback(
    Output("view-recipe-add-command-json-editor", "value"),
    Input("view-recipe-add-command-command-dropdown", "value"),
    [
        State("url", "pathname"),
        State("view-recipe-add-command-device-dropdown", "value"),
    ],
    prevent_initial_call=True,
)
def view_recipe_fill_add_command_json_editor(command_type, url, device_type):
    if url == "/view-recipe":
        if command_type is None or command_type == "":
            return ""
        args_dict = {}
        args_list = util.devices_ref_redundancy[device_type]["commands"][command_type][
            "args"
        ]
        for arg in args_list:
            args_dict[arg] = args_list[arg]["default"]
        args_dict["delay"] = 0.0
        return json.dumps(args_dict, indent=4)


@app.callback(
    [
        Output("view-recipe-add-command-editor", "disabled"),
        Output("add-command-error", "children"),
    ],
    Input("view-recipe-add-command-json-editor", "value"),
    State("url", "pathname"),
)
def view_recipe_check_add_command_json(value, url):
    if str(url) == "/view-recipe":
        if value == "" or value is None:
            return True, []
        try:
            json.loads(value)
            return False, []
        except Exception as e:
            return True, ["Invalid JSON: " + str(e)]


@app.callback(
    [
        Output("view-recipe-command-add-modal", "is_open", allow_duplicate=True),
        Output("view-recipe-alert", "is_open", allow_duplicate=True),
        Output("view-recipe-alert", "children", allow_duplicate=True),
        Output("view-recipe-alert", "color", allow_duplicate=True),
        Output("view-recipe-alert", "duration", allow_duplicate=True),
        Output("commands-table", "data", allow_duplicate=True),
    ],
    Input("view-recipe-add-command-editor", "n_clicks"),
    [
        State("url", "pathname"),
        State("view-recipe-add-command-device-dropdown", "value"),
        State("view-recipe-add-command-command-dropdown", "value"),
        State("view-recipe-add-command-json-editor", "value"),
    ],
    prevent_initial_call=True,
)
def view_recipe_add_command(n, url, device_type, command_type, json_value):
    if str(url) == "/view-recipe":
        if json_value == "" or json_value is None:
            return [False, True, ["Something went wrong"], "danger", 3000]
        try:
            com.add_command_from_dict(device_type, command_type, json.loads(json_value))
            update_upstream_recipe_dict()
            return [False, True, ["Command added"], "success", 3000, None]
        except Exception as e:
            print(str(e))
            return [
                False,
                True,
                ["Something went wrong: " + str(e)],
                "danger",
                3000,
                None,
            ]


@app.callback(
    [
        Output("view-recipe-execution-options-output-files", "value"),
        Output("view-recipe-execution-options-default-execution-record-name", "value"),
    ],
    Input("url", "pathname"),
)
def view_recipe_fill_execution_options(url):
    if str(url) == "/view-recipe":
        if "document" in com.__dict__.keys():
            ls = com.document["recipe_dict"]["execution_options"]["output_files"]
            filesToRet = ""
            for item in ls:
                filesToRet += item + "\n"
            return [
                filesToRet,
                com.document["recipe_dict"]["execution_options"][
                    "default_execution_record_name"
                ],
            ]
        return ["", ""]


@app.callback(
    [
        Output("view-recipe-execution-options-saved-label", "children"),
        Output("view-recipe-execution-options-saved-label", "style"),
    ],
    Input("view-recipe-execution-options-save-button", "n_clicks"),
    [
        State("view-recipe-execution-options-output-files", "value"),
        State("view-recipe-execution-options-default-execution-record-name", "value"),
        State("url", "pathname"),
    ],
    prevent_initial_call=True,
)
def view_recipe_save_execution_options(
    n, filenames, default_execution_record_name, url
):
    if str(url) == "/view-recipe":
        com.execution_options["output_files"] = filenames.splitlines()
        com.execution_options[
            "default_execution_record_name"
        ] = default_execution_record_name
        success = update_upstream_recipe_dict()
        if success:
            return ["Saved!", {"display": "block"}]
        else:
            return ["Something went wrong", {"display": "block"}]
    return ["", {"display": "none"}]


# ---------------------------------------------------
# Python Edit Recipe Page
# ---------------------------------------------------


@app.callback(
    [
        Output("ace-recipe-editor", "value", allow_duplicate=True),
        Output("ace-editor-alert", "is_open"),
        Output("ace-editor-alert", "children"),
        Output("ace-editor-alert", "color"),
        Output("ace-editor-alert", "duration"),
    ],
    [Input("refresh-button-ace", "n_clicks")],
    State("url", "pathname"),
    # prevent_initial_call=True,
)
def fill_ace_editor(n, url):  # python-edit-recipe page
    if url == "/edit-recipe":
        if hasattr(com, "document"):
            python_code = com.document.get("python_code", "")
            if python_code is not None and python_code != "":
                print("fill_ace_editor")
                return [str(python_code), True, "Loaded!", "success", 1500]
            else:
                print("fill_ace_editor")
                return ["", True, "No code available", "warning", 1000]
        else:
            return ["", True, "No code available", "warning", 1000]
    else:
        return ["", False, "na", "success", 0]


@app.callback(
    [
        Output("ace-recipe-editor", "value", allow_duplicate=True),
        Output("ace-editor-alert", "is_open", allow_duplicate=True),
        Output("ace-editor-alert", "children", allow_duplicate=True),
        Output("ace-editor-alert", "color", allow_duplicate=True),
        Output("ace-editor-alert", "duration", allow_duplicate=True),
    ],
    Input("execute-and-save-button", "n_clicks"),
    [State("ace-recipe-editor", "value")],
    prevent_initial_call=True,
)
def execute_and_save(n, value):  # python-edit-recipe page
    print("execute_and_save")
    if value is not None and value != "":
        try:
            exec(value)
            doc_id = com.document.get("_id", "")
            (mongo.update_yaml_file("recipes", doc_id, {"python_code": value}))
            com.load_from_yaml("to_save.yaml")
            com.document = mongo.find_documents("recipes", {"_id": doc_id})[0]
            return [str(value), True, "Saved!", "success", 1500]
        except Exception as e:
            return [str(value), True, str(e), "danger", 5000]
    else:
        return ["", True, "No code to execute", "warning", 1000]

    # return ["", False, "", "success", 1500]


@app.callback(
    [
        Output("add-device-dropdown-ace", "options"),
        Output("add-device-dropdown-ace", "value"),
    ],
    [Input("device-add-modal-ace", "is_open")],
    prevent_initial_call=True,
)
def fill_device_add_modal_ace(is_open):  # python-edit-recipe page
    if is_open:
        print("fill_device_add_modal_ace")
        return list(util.devices_ref_redundancy.keys()), ""
    return [], ""


@app.callback(
    [
        Output("add-command-device-dropdown-ace", "options"),
        Output("add-command-device-dropdown-ace", "value"),
    ],
    [Input("command-add-modal-ace", "is_open")],
    prevent_initial_call=True,
)
def fill_command_device_add_modal_ace(is_open):  # python-edit-recipe page
    if is_open:
        print("fill_command_device_add_modal_ace")
        return list(util.devices_ref_redundancy.keys()), ""
    return [], ""


@app.callback(
    [
        Output("add-command-command-dropdown-ace", "options"),
        Output("add-command-command-dropdown-ace", "value"),
    ],
    [Input("add-command-device-dropdown-ace", "value")],
    prevent_initial_call=True,
)
def fill_command_add_modal_ace(device):  # python-edit-recipe page
    if device is not None and device != "":
        print("fill_command_add_modal_ace")
        return list(util.devices_ref_redundancy[device]["commands"].keys()), ""
    return [], ""


@app.callback(
    Output("add-device-editor-ace", "disabled"),
    [
        Input("add-device-dropdown-ace", "value"),
        Input("device-add-modal-ace", "is_open"),
    ],
    State("device-add-modal-ace", "is_open"),
    prevent_initial_call=True,
)
def enable_add_device_button_ace(value, is_openInp, is_open):  # python-edit-recipe page
    if value == "" or value is None:
        return True
    print("enable_add_device_button_ace")
    return False


@app.callback(
    Output("add-command-editor-ace", "disabled"),
    [
        Input("add-command-command-dropdown-ace", "value"),
        Input("command-add-modal-ace", "is_open"),
    ],
    State("command-add-modal-ace", "is_open"),
    prevent_initial_call=True,
)
def enable_add_command_button_ace(
    value, is_openInp, is_open
):  # python-edit-recipe page
    if value == "" or value is None:
        return True
    print("enable_add_command_button_ace")
    return False


@app.callback(
    [
        Output("ace-recipe-editor", "value"),
        Output("ace-editor-alert", "is_open", allow_duplicate=True),
        Output("ace-editor-alert", "children", allow_duplicate=True),
        Output("ace-editor-alert", "color", allow_duplicate=True),
        Output("ace-editor-alert", "duration", allow_duplicate=True),
    ],
    Input("add-device-editor-ace", "n_clicks"),
    [State("ace-recipe-editor", "value"), State("add-device-dropdown-ace", "value")],
    prevent_initial_call=True,
)
def add_device_to_recipe_ace(n_clicks, value, device_type):  # python-edit-recipe page
    print("add_device_to_recipe_ace")
    if value == "" or value is None:
        return ["", True, "No code in editor", "warning", 3000]
    try:
        value = str(value)
        import_line = util.devices_ref_redundancy[device_type]["import_device"]
        init_line = util.devices_ref_redundancy[device_type]["init"]["default_code"]
        if import_line not in value:
            value = import_line + "\n" + value
        value = value.replace(
            "##################################################\n##### Add commands to the command sequence",
            "seq.add_device("
            + init_line
            + ")\n\n##################################################\n##### Add commands to the command sequence",
        )
        return [str(value), True, "Device added successfully", "success", 3000]
    except Exception as e:
        print(e)
        return [str(value), True, "Error adding device: " + str(e), "danger", 6000]


@app.callback(
    [
        Output("ace-recipe-editor", "value", allow_duplicate=True),
        Output("ace-editor-alert", "is_open", allow_duplicate=True),
        Output("ace-editor-alert", "children", allow_duplicate=True),
        Output("ace-editor-alert", "color", allow_duplicate=True),
        Output("ace-editor-alert", "duration", allow_duplicate=True),
    ],
    Input("add-command-editor-ace", "n_clicks"),
    [
        State("ace-recipe-editor", "value"),
        State("add-command-command-dropdown-ace", "value"),
        State("add-command-device-dropdown-ace", "value"),
    ],
    prevent_initial_call=True,
)
def add_commands_to_recipe_ace(
    n_clicks, value, command, device_type
):  # python-edit-recipe page
    print("add_commands_to_recipe_ace")
    og_value = str(value)
    if value == "" or value is None:
        return ["", True, "No code in editor", "warning", 3000]
    try:
        value = str(value)
        command_line = util.devices_ref_redundancy[device_type]["commands"][command][
            "default_code"
        ]
        import_line = util.devices_ref_redundancy[device_type]["import_commands"]
        import_device_line = util.devices_ref_redundancy[device_type]["import_device"]
        if import_device_line not in value:
            raise Exception(
                "Device (or its import '"
                + import_device_line
                + "') not found in recipe"
            )
        if import_line not in value:
            value = import_line + "\n" + value
        if "\nrecipe_file = 'to_save.yaml'\nseq.save_to_yaml(recipe_file)" not in value:
            raise Exception("Code is not in valid format")
        value = value.replace(
            "\nrecipe_file = 'to_save.yaml'\nseq.save_to_yaml(recipe_file)",
            "\nseq.add_command("
            + command_line
            + ")\n\n\nrecipe_file = 'to_save.yaml'\nseq.save_to_yaml(recipe_file)",
        )
        return [str(value), True, "Command added successfully", "success", 3000]
    except Exception as e:
        print(e)
        return [og_value, True, str("Error adding command: " + str(e)), "danger", 6000]


# ---------------------------------------------------
# Execute Recipe Page
# ---------------------------------------------------


def kill_execution():  # execute-recipe page
    os.kill(os.getpid(), signal.SIGINT)


@app.callback(
    Output("hidden-div", "children"),
    Input("stop-button", "n_clicks"),
    prevent_initial_call=True,
)
def stop_execution(n):  # execute-recipe page
    print("stop_execution")
    kill_execution()
    return []


@app.callback(
    Output("execute-recipe-output", "children"),
    [Input("execute-button", "n_clicks")],
    prevent_initial_call=True,
)
def execute_recipe(n_clicks):  # execute-recipe page
    print("execute_recipe")
    invoker.invoking = True
    invoker.invoke_commands()
    invoker.invoking = False
    return ["done"]


@app.callback(
    Output("console-out2", "children"),
    Input("interval1", "n_intervals"),
    [State("url", "pathname")],
    prevent_initial_call=True,
)
def update_output(n, url):  # execute-recipe page
    if url == "/execute-recipe":
        log_string = ""
        log_list = invoker.get_log_messages()
        for msg in log_list:
            log_string += msg
            # log_string += '<br>'
        return html.Pre(log_string)
    return ""


@app.callback(
    Output("console-out2", "children", allow_duplicate=True),
    Input("reset-button", "n_clicks"),
    prevent_initial_call=True,
)
def reset_console(n):  # execute-recipe page
    print("reset_console")
    invoker.clear_log_file()
    # dashLoggerHandler.queue = []
    return []


@app.callback(
    [
        Output("execute-recipe-upload-document", "value"),
        Output("execute-recipe-upload-name", "value"),
    ],
    Input("reset-button", "n_clicks"),
    State("url", "pathname"),
)
def execute_recipe_load_document_viewer(n, url):
    if str(url) == "/execute-recipe":
        if "document" in list(com.__dict__.keys()):
            if com.device_list != []:
                toRet = ""
                recipe_ec = com.get_recipe()
                for device in recipe_ec[0]:
                    toRet += str(device) + "\n"
                toRet += "\n"
                for command in recipe_ec[1]:
                    toRet += str(command) + "\n"
                if "default_execution_record_name" in list(
                    com.execution_options.keys()
                ):
                    return [
                        toRet,
                        com.execution_options["default_execution_record_name"],
                    ]
                return [toRet, "Execution"]
        return ["", "No Recipe Loaded"]


@app.callback(
    [Output("execute-recipe-upload-data-output", "children")],
    Input("execute-recipe-upload-data-button", "n_clicks"),
    [
        State("url", "pathname"),
        State("execute-recipe-upload-name", "value"),
        State("execute-recipe-upload-document", "value"),
        State("console-out2", "children"),
        State("execute-recipe-upload-notes", "value"),
        State("execute-recipe-upload-files", "contents"),
        State("execute-recipe-upload-files", "filename"),
    ],
    prevent_initial_call=True,
)
def execute_recipe_upload_data(
    n_clicks, url, name, recipe_data, console_log, notes, files, filenames
):
    if str(url) == "/execute-recipe":
        print("execute_recipe_upload_data")
        execution = {}
        execution["name"] = name
        if (
            isinstance(recipe_data, list)
            and recipe_data is not None
            and recipe_data != []
        ):
            execution["recipe"] = recipe_data[0].split("\n")
        elif recipe_data is not None and recipe_data != "":
            execution["recipe"] = recipe_data.split("\n")
        execution["notes"] = notes
        execution["log"] = str(console_log["props"]["children"]).split("\n")
        execution["files"] = []
        if isinstance(files, list):
            for i, file in enumerate(files):
                # print(file)
                file_bytes = base64.b64decode(file + "==")
                execution["files"].append(
                    mongo_gridfs.put(file_bytes, filename=filenames[i])
                )
        exec_success = update_execution_upstream(execution)
        # exec_success = False
        if exec_success:
            return [html.P("Data uploaded successfully")]
        else:
            return [html.P("Error uploading data")]


# ---------------------------------------------------
# Data Page
# ---------------------------------------------------


def render_dict(data):  # data (and maybe database) page
    if isinstance(data, dict):
        return [
            dbc.Accordion(
                [
                    dbc.AccordionItem(
                        [
                            html.Div(
                                render_dict(value),
                                style={"margin-left": "15px"},
                            )
                        ],
                        title=key,
                    )
                    for key, value in data.items()
                ]
            )
        ]
    else:
        return html.P(str(data))


@app.callback(
    Output("data-output-div2", "children"),
    Input("load-data-button", "n_clicks"),
    prevent_initial_call=True,
)
def load_data_accordion(n):  # data page
    if not hasattr(com, "document"):
        print("load_data_accordion - no document")
        return []
    print("load_data_accordion")
    return render_dict(com.document)


# ---------------------------------------------------
# Manual Control Recipe Page
# ---------------------------------------------------


@app.callback(
    Output("manual-control-device-dropdown", "options"),
    Input("interval-manual-control", "n_intervals"),
    State("url", "pathname"),
)
def fill_manual_control_device_dropdown(n, url):  # manual-control page
    if str(url) == "/manual-control":
        print("fill_manual_control_device_dropdown")
        return list(util.devices_ref_redundancy.keys())


@app.callback(
    [
        Output("manual-control-command-dropdown", "disabled"),
        Output("manual-control-command-dropdown", "options"),
    ],
    Input("manual-control-device-dropdown", "value"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def fill_manual_control_command_dropdown(val, url):  # manual-control page
    if str(url) == "/manual-control":
        print("fill_manual_control_command_dropdown")
        if val is None or val == "":
            return [True, []]
        else:
            if util.devices_ref_redundancy[val]["serial"] == False:
                return [
                    False,
                    list(util.devices_ref_redundancy[val]["commands"].keys()),
                ]
            else:
                toRet = list(util.devices_ref_redundancy[val]["commands"].keys()).copy()
                for command in util.devices_ref_redundancy[val]["serial_sequence"]:
                    if command in toRet:
                        toRet.remove(command)
                return [False, toRet]


@app.callback(
    [Output("manual-control-device-form", "children")],
    Input("manual-control-device-dropdown", "value"),
    State("url", "pathname"),
)
def create_manual_control_device_form(value, url):
    if str(url) == "/manual-control":
        print("create_manual_control_device_form")
        if value is None or value == "":
            return [[]]
        else:
            args = util.devices_ref_redundancy[value]["init"]["args"]
            toRet = []
            for arg in args:
                if arg == "port":
                    toRet.append(
                        dbc.Row(
                            [
                                dbc.Label(
                                    dbc.NavLink(
                                        arg,
                                        n_clicks=0,
                                        id="manual-control-port-field",
                                        style={
                                            "cursor": "pointer",
                                            "color": "blue",
                                            "textDecoration": "underline",
                                        },
                                    ),
                                    html_for=str(value + "+" + arg),
                                    width=2,
                                ),
                                dbc.Col(
                                    [
                                        dbc.Input(
                                            id=str(value + "+" + arg),
                                            value=args[arg]["default"],
                                            placeholder=args[arg]["notes"],
                                        ),
                                    ],
                                    width=10,
                                ),
                            ],
                            className="mb-2",
                        )
                    )
                else:
                    toRet.append(
                        dbc.Row(
                            [
                                dbc.Label(
                                    [arg], html_for=str(value + "+" + arg), width=2
                                ),
                                dbc.Col(
                                    [
                                        dbc.Input(
                                            id=str(value + "+" + arg),
                                            value=args[arg]["default"],
                                            placeholder=args[arg]["notes"],
                                        ),
                                    ],
                                    width=10,
                                ),
                            ],
                            className="mb-2",
                        )
                    )

            return [toRet]


@app.callback(
    [Output("manual-control-command-form", "children")],
    [
        Input("manual-control-command-dropdown", "value"),
        Input("manual-control-device-dropdown", "value"),
    ],
    [
        State("url", "pathname"),
        State("manual-control-device-form", "children"),
    ],
)
def create_manual_control_command_form(command, device, url, device_form):
    if str(url) == "/manual-control":
        print("create_manual_control_command_form")
        if command is None or command == "" or device is None or device == "":
            return [[]]
        else:
            toRet = []
            if util.devices_ref_redundancy[device]["serial"] == True:
                seq_toRet = []
                for seq_command in util.devices_ref_redundancy[device][
                    "serial_sequence"
                ]:
                    seq_toRet.append(dbc.Row([dbc.Label([seq_command])]))
                    args = util.devices_ref_redundancy[device]["commands"][seq_command][
                        "args"
                    ]
                    for arg in args:
                        seq_toRet.append(
                            dbc.Row(
                                [
                                    dbc.Label(
                                        [arg],
                                        html_for=str(
                                            device + "+" + seq_command + "+" + arg
                                        ),
                                        width=2,
                                    ),
                                    dbc.Col(
                                        [
                                            dbc.Input(
                                                id=str(
                                                    device
                                                    + "+"
                                                    + seq_command
                                                    + "+"
                                                    + arg
                                                ),
                                                value=args[arg]["default"],
                                                placeholder=args[arg]["notes"],
                                            ),
                                        ],
                                        width=10,
                                    ),
                                ],
                                className="mb-2",
                            )
                        )
                toRet.append(dbc.Row(seq_toRet, className="mb-3"))
            toRet.append(dbc.Row([dbc.Label([command])]))
            args = util.devices_ref_redundancy[device]["commands"][command]["args"]
            com_toRet = []
            for arg in args:
                com_toRet.append(
                    dbc.Row(
                        [
                            dbc.Label(
                                [arg],
                                html_for=str(device + command + "+" + arg),
                                width=2,
                            ),
                            dbc.Col(
                                [
                                    dbc.Input(
                                        id=str(device + command + "+" + arg),
                                        value=args[arg]["default"],
                                        placeholder=args[arg]["notes"],
                                    ),
                                ],
                                width=10,
                            ),
                        ],
                        className="mb-2",
                    )
                )
            toRet.append(dbc.Row(com_toRet, className="mb-3"))
            return [toRet]


@app.callback(
    Output("manual-control-device-dropdown", "value"),
    Input("manual-control-clear-button", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def manual_control_clear_form(n, url):
    if str(url) == "/manual-control":
        print("manual_control_clear_form")
        return None


@app.callback(
    Output("manual-control-clear-button", "disabled"),
    Input("manual-control-device-dropdown", "value"),
    State("url", "pathname"),
)
def manual_control_clear_button(value, url):
    if str(url) == "/manual-control":
        print("manual_control_clear_button")
        if value is None or value == "":
            return True
        else:
            return False


@app.callback(
    Output("manual-control-open-execute-modal-button", "disabled"),
    Input("manual-control-command-dropdown", "value"),
    State("url", "pathname"),
)
def manual_control_execute_button(value, url):
    if str(url) == "/manual-control":
        print("manual_control_execute_button")
        if value is None or value == "":
            return True
        else:
            return False


@app.callback(
    [
        Output("manual-control-execute-modal", "is_open", allow_duplicate=True),
        Output("manual-control-execute-modal-body", "children", allow_duplicate=True),
    ],
    Input("manual-control-open-execute-modal-button", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def open_manual_control_execute_modal(n, url):
    if str(url) == "/manual-control":
        print("open_manual_control_execute_modal")
        return True, []


@app.callback(
    [
        Output("manual-control-command-dropdown", "className"),
        Output("manual-control-alert", "is_open", allow_duplicate=True),
        Output("manual-control-alert", "children", allow_duplicate=True),
        Output("manual-control-alert", "color", allow_duplicate=True),
        Output("manual-control-alert", "duration", allow_duplicate=True),
        Output(
            "manual-control-execute-modal-body-code", "children", allow_duplicate=True
        ),
    ],
    Input("manual-control-open-execute-modal-button", "n_clicks"),
    [
        State("url", "pathname"),
        State("manual-control-command-dropdown", "className"),
        State("manual-control-device-dropdown", "value"),
        State("manual-control-command-dropdown", "value"),
        State("manual-control-device-form", "children"),
        State("manual-control-command-form", "children"),
    ],
    prevent_initial_call=True,
)
def manual_control_execute_fill_code(
    n, url, opt, device, command, device_form, command_form
):
    if str(url) == "/manual-control":
        print("manual_control_execute_fill_code")
        if device is None or device == "" or command is None or command == "":
            return (
                opt,
                True,
                "Something went wrong. Check the fields below.",
                "danger",
                3000,
            )
        code = ""
        code += util.devices_ref_redundancy[device]["import_device"]
        code += "\n"
        code += util.devices_ref_redundancy[device]["import_commands"]
        code += "\n"
        code += "from command_sequence import CommandSequence\nfrom command_invoker import CommandInvoker\n"

        instantiate_code = ""
        instantiate_code += util.devices_ref_redundancy[device]["init"]["obj_name"]
        instantiate_code += "("
        for i, arg in enumerate(util.devices_ref_redundancy[device]["init"]["args"]):
            if i != 0:
                instantiate_code += ", "
            instantiate_code += arg + "="
            if util.devices_ref_redundancy[device]["init"]["args"][arg]["type"] == str:
                instantiate_code += "'"
                instantiate_code += str(
                    device_form[i]["props"]["children"][1]["props"]["children"][0][
                        "props"
                    ]["value"]
                )
                instantiate_code += "'"
            else:
                instantiate_code += str(
                    device_form[i]["props"]["children"][1]["props"]["children"][0][
                        "props"
                    ]["value"]
                )
        instantiate_code += ")"
        code += str(device) + "_seq" + " = CommandSequence()"
        code += "\n\n"
        code += str(device) + "_seq" + ".add_device(" + instantiate_code + ")"
        code += "\n"

        if util.devices_ref_redundancy[device]["serial"] == True:
            for i, serial_seq_command in enumerate(
                util.devices_ref_redundancy[device]["serial_sequence"]
            ):
                code += (
                    str(device)
                    + "_seq"
                    + ".add_command("
                    + str(serial_seq_command)
                    + "("
                )
                for ii, serial_seq_command_arg in enumerate(
                    util.devices_ref_redundancy[device]["commands"][serial_seq_command][
                        "args"
                    ]
                ):
                    if ii != 0:
                        code += ", "
                    if serial_seq_command_arg == "receiver":
                        code += (
                            serial_seq_command_arg
                            + "="
                            + str(device)
                            + "_seq.device_by_name['"
                            + str(
                                command_form[0]["props"]["children"][(2 * ii) + 1][
                                    "props"
                                ]["children"][1]["props"]["children"][0]["props"][
                                    "value"
                                ]
                            )
                            + "']"
                        )
                    elif (
                        util.devices_ref_redundancy[device]["commands"][
                            serial_seq_command
                        ]["args"][serial_seq_command_arg]["type"]
                        == str
                    ):
                        code += (
                            serial_seq_command_arg
                            + "="
                            + "'"
                            + str(
                                command_form[0]["props"]["children"][(2 * ii) + 1][
                                    "props"
                                ]["children"][1]["props"]["children"][0]["props"][
                                    "value"
                                ]
                            )
                            + "'"
                        )
                    else:
                        code += (
                            serial_seq_command_arg
                            + "="
                            + str(
                                command_form[0]["props"]["children"][(2 * ii) + 1][
                                    "props"
                                ]["children"][1]["props"]["children"][0]["props"][
                                    "value"
                                ]
                            )
                        )

                code += "))\n"
            code += str(device) + "_seq" + ".add_command(" + str(command) + "("
            for ii, seq_command_arg in enumerate(
                util.devices_ref_redundancy[device]["commands"][command]["args"]
            ):
                if ii != 0:
                    code += ", "
                if seq_command_arg == "receiver":
                    code += (
                        seq_command_arg
                        + "="
                        + str(device)
                        + "_seq.device_by_name['"
                        + str(
                            command_form[2]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                        + "']"
                    )
                elif (
                    util.devices_ref_redundancy[device]["commands"][command]["args"][
                        seq_command_arg
                    ]["type"]
                    == str
                ):
                    code += (
                        seq_command_arg
                        + "="
                        + "'"
                        + str(
                            command_form[2]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                        + "'"
                    )
                else:
                    code += (
                        seq_command_arg
                        + "="
                        + str(
                            command_form[2]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                    )

            code += "))\n\n"
        else:
            code += str(device) + "_seq" + ".add_command(" + str(command) + "("
            for ii, seq_command_arg in enumerate(
                util.devices_ref_redundancy[device]["commands"][command]["args"]
            ):
                if ii != 0:
                    code += ", "
                if seq_command_arg == "receiver":
                    code += (
                        seq_command_arg
                        + "="
                        + str(device)
                        + "_seq.device_by_name['"
                        + str(
                            command_form[1]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                        + "']"
                    )
                elif (
                    util.devices_ref_redundancy[device]["commands"][command]["args"][
                        seq_command_arg
                    ]["type"]
                    == str
                ):
                    code += (
                        seq_command_arg
                        + "="
                        + "'"
                        + str(
                            command_form[1]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                        + "'"
                    )
                else:
                    code += (
                        seq_command_arg
                        + "="
                        + str(
                            command_form[1]["props"]["children"][ii]["props"][
                                "children"
                            ][1]["props"]["children"][0]["props"]["value"]
                        )
                    )

            code += "))\n\n"
        code += (
            str(device)
            + "_seq_invoker = CommandInvoker("
            + str(device)
            + "_seq, False, False, False)\n"
        )
        code += str(device) + "_seq_invoker.invoke_commands()"
        interceptor = ConsoleInterceptor()
        print("\n")
        interceptor.start_interception()
        print(code)
        interceptor.stop_interception()
        code_output = interceptor.get_intercepted_messages()
        del interceptor
        code_log_string = ""
        for msg in code_output:
            code_log_string += msg
        # interceptor = ConsoleInterceptor()
        # interceptor.start_interception()
        # try:
        # exec(code)
        # interceptor.stop_interception()
        # messages = interceptor.get_intercepted_messages()
        # log_string = ""
        # for msg in messages:
        #     if msg == '\r':
        #         log_string += '\n'
        #     else:
        #         log_string += msg
        # print(messages)
        return (
            opt,
            True,
            "Command(s) ready to execute.",
            "warning",
            0,
            html.Pre(code_log_string),
        )
        # except Exception as e:
        #     print(e)
        #     # interceptor.stop_interception()
        #     # messages = interceptor.get_intercepted_messages()
        #     # log_string = ""
        #     # for msg in messages:
        #     #     log_string += msg
        #     return opt, True, "Something went wrong. Check logs.", "Warning", 0, True, html.Pre(code_log_string)

    return opt


@app.callback(
    [
        Output("manual-control-alert", "is_open"),
        Output("manual-control-alert", "children"),
        Output("manual-control-alert", "color"),
        Output("manual-control-alert", "duration"),
        Output("manual-control-execute-modal-body", "children"),
    ],
    [Input("manual-control-execute-button", "n_clicks")],
    [
        State("url", "pathname"),
        State("manual-control-execute-modal-body-code", "children"),
    ],
    prevent_initial_call=True,
)
def manual_control_execute_code(n, url, code):
    if str(url) == "/manual-control":
        print("manual_control_execute_code")
        interceptor = ConsoleInterceptor()
        # print('\n')
        # interceptor.start_interception()
        # print(code)
        # interceptor.stop_interception()
        # code_output = interceptor.get_intercepted_messages()
        # code_log_string = ""
        # for msg in code_output:
        #     code_log_string += msg
        # interceptor = ConsoleInterceptor()
        interceptor.start_interception()
        try:
            exec(code["props"]["children"])
            interceptor.stop_interception()
            messages = interceptor.get_intercepted_messages()
            del interceptor
            log_string = ""
            for msg in messages:
                if msg == "\r":
                    log_string += "\n"
                else:
                    log_string += msg
            return True, "Execution complete.", "success", 0, html.Pre(log_string)
        except Exception as e:
            print(e)
            interceptor.stop_interception()
            messages = interceptor.get_intercepted_messages()
            del interceptor
            log_string = ""
            for msg in messages:
                log_string += msg
            return (
                True,
                "Something went wrong. Check logs.",
                "danger",
                0,
                html.Pre(log_string),
            )
    return False, "", "success", 0, html.Pre("")


@app.callback(
    [
        Output("manual-control-port-modal", "is_open"),
        Output("manual-control-serial-ports-info", "children"),
    ],
    Input("manual-control-port-field", "n_clicks"),
    prevent_initial_call=True,
)
def open_fill_manual_control_serial(n):
    if _has_serial and n != 0:
        ports = serial.tools.list_ports.comports()
        str_ports = ""
        for port, desc, hwid in sorted(ports):
            str_ports += f"{port}: {desc} [{hwid}]\n"
        lines = str_ports.splitlines()
        return True, [html.Div([html.Div(line) for line in lines])]
    return False, []


# ---------------------------------------------------------------
# Database page
# ---------------------------------------------------------------


@app.callback(
    Output("database-db-dropdown", "options"),
    Input("interval-database", "n_intervals"),
    State("url", "pathname"),
)
def fill_database_db_dropdown(n, url):
    if str(url) == "/database":
        print("fill_database_db_dropdown")
        return list(mongo.client.list_database_names())


@app.callback(
    [
        Output("database-collection-dropdown", "disabled"),
        Output("database-collection-dropdown", "options"),
    ],
    [Input("database-db-dropdown", "value")],
    State("url", "pathname"),
    prevent_initial_call=True,
)
def fill_database_collection_dropdown(db, url):  # database page
    if str(url) == "/database":
        print("fill_database_collection_dropdown")
        if db is not None and db != "":
            return False, list(mongo.client[db].list_collection_names())
    return True, []


@app.callback(
    [
        Output("database-document-dropdown", "disabled"),
        Output("database-document-dropdown", "options"),
    ],
    Input("database-collection-dropdown", "value"),
    [State("url", "pathname"), State("database-db-dropdown", "value")],
    prevent_initial_call=True,
)
def fill_database_document_dropdown(collection, url, db):
    if str(url) == "/database":
        if collection is not None and collection != "":
            print("fill_database_document_dropdown")
            docs = list(mongo.client[db][collection].find({}))
            toRet = []
            for doc in docs:
                toRet.append(str(doc["_id"]))
            return False, toRet
        else:
            return True, []


def process_schema(schema):
    if isinstance(schema, dict):
        if "bsonType" not in list(schema.keys()):
            for key in list(schema.keys()):
                schema[key] = process_schema(schema[key])
        elif "bsonType" in list(schema.keys()):
            schema["Properties"] = {"Data Type": schema["bsonType"]}
            del schema["bsonType"]
            if "description" in list(schema.keys()):
                schema["Properties"].update({"Description": schema["description"]})
                del schema["description"]
            if "required" in list(schema.keys()):
                # schema['Properties'].update({"Required": schema['required']})
                del schema["required"]
            if "properties" in list(schema.keys()):
                schema["Variables"] = process_schema(schema["properties"])
                del schema["properties"]
            if "title" in list(schema.keys()):
                del schema["title"]

    return schema


@app.callback(
    Output("database-collection-schema", "children"),
    [
        Input("database-collection-dropdown", "value"),
        Input("database-db-dropdown", "value"),
    ],
    [State("url", "pathname")],
    prevent_initial_call=True,
)
def fill_database_collection_schema(collection, db, url):
    if str(url) == "/database":
        print("fill_database_collection_schema")
        if collection is not None and collection != "" and db is not None and db != "":
            try:
                schema = (
                    mongo.client[db].get_collection(collection).options()["validator"]
                )
                schema = process_schema(schema)
                return render_dict(schema)
            except Exception as e:
                return ["Validation rules missing or something went wrong."]
    return []


@app.callback(
    Output("database-document-viewer", "children"),
    [
        Input("database-document-dropdown", "value"),
        Input("database-collection-dropdown", "value"),
    ],
    [State("url", "pathname"), State("database-db-dropdown", "value")],
    prevent_initial_call=True,
)
def fill_database_document_viewer(document, collection, url, db):
    if str(url) == "/database":
        print("fill_database_document_viewer")
        if document is not None and document != "" and db is not None and db != "":
            # try:
            doc = mongo.client[db][collection].find({"_id": ObjectId(document)})[0]
            return render_dict(doc)
            # except Exception as e:
            #     return ["Document missing or something went wrong"]
    return []


# ---------------------------------------------------------------
# Real Time Telemetry page
# ---------------------------------------------------------------


@app.callback(
    [Output("real-time-telemetry-div", "children")],
    [Input("interval-real-time-telemetry", "n_intervals")],
    [Input("real-time-telemetry-device-dropdown", "value"), State("url", "pathname")],
    prevent_initial_call=True,
)
def fill_real_time_telemetry(n, device, url):
    if str(url) == "/real-time-telemetry" and device is not None and device != "":
        print("fill_real_time_telemetry")
        if "telemetry" in list(util.devices_ref_redundancy[device].keys()):
            device_parameter_options = util.devices_ref_redundancy[device]["telemetry"][
                "options"
            ]
            for parameter in util.devices_ref_redundancy[device]["telemetry"]:
                print(parameter)
                parameter_options = util.devices_ref_redundancy[device]["telemetry"][
                    "parameters"
                ][parameter]

        return [""]
    return [""]


if __name__ == "__main__":
    app.run(debug=False)


# @app.callback(
#     Output("data-output-div", "children"),
#     Input("load-data-button", "n_clicks"),
#     prevent_initial_call=True,
# )
# def load_data(n):
#     print('load_data')
#     # val = ""
#     # docs = mongo.db["recipes"].find()
#     # for doc in docs:
#     #     val += str(doc)
#     #     val += "<br><br>"
#     # nval = val.
#     return (json.dumps(str(com.document)))


# class DashLoggerHandler(logging.StreamHandler):
#     def __init__(self):
#         logging.StreamHandler.__init__(self)
#         self.queue = []

#     def emit(self, record):
#         msg = self.format(record)
#         self.queue.append(msg)


# logger = logging.getLogger()
# logger.setLevel(logging.DEBUG)
# dashLoggerHandler = DashLoggerHandler()
# logger.addHandler(dashLoggerHandler)

# logger = logging.getLogger(invoker.log.name)
# logger.setLevel(logging.DEBUG)
# from io import StringIO
# log_capture = StringIO()

# # Create a stream handler and set its stream to the log_capture object
# stream_handler = logging.StreamHandler(log_capture)
# logger.addHandler(stream_handler)
# log_messages = []

# import sys
# from io import StringIO
# stringio = StringIO()
# sys.stdout = stringio
# @app.callback(Output('console-output', 'value'), [Input('update-interval', 'n_intervals')])
# def show_console_output(n):
#     # Retrieve console output from StringIO object
#     stringio.seek(0)
#     console_output = stringio.read()
#     return console_output

# @app.callback(Output("page-content", "children"), [Input("url", "pathname")])
# def display_page(pathname):
#     print("\n\nrefreshing to " + pathname)
#     if pathname == "/":
#         return home_layout
#     elif pathname == "/edit-recipe":
#         return edit_recipe_layout
#     elif pathname == "/execute-recipe":
#         return execute_recipe_layout
#     elif pathname == "/data":
#         return data_layout
#     else:
#         return html.Div("404")


# data_list4 = com.device_list

# dl5 = []
# for list in data_list4:
#     dl5.append(list.__dict__)

# print(dl5[2]['motor'].__dict__)


# @app.callback(
#         Output("commands-accordion", "children"),
#         Input("add-command-button", "n_clicks"),
#         State("commands-accordion", "children"),
#         prevent_initial_call=True,
#         allow_duplicate=True
# )
# def add_command_accordian(n_clicks, children):
#     children=[
#         dbc.AccordionItem(
#             "new item", title="new title", item_id="new")
#     ]

#     return children


# @app.callback(
#     Output("commands-accordion", "children"),
#     Input("refresh-button2", "n_clicks"),
#     State("commands-accordion", "children"),
#     # allow_duplicate=True
# )
# def load_commands_accordion(n_clicks, children):
#     children = []
#     # print()
#     command_list = com.get_unlooped_command_list().copy()
#     # print(command_list)
#     # for command in command_list:
#     #     if isinstance(command, CompositeCommand):
#     #         for sub_command in command._command_list:
#     #             sub_command._receiver = sub_command._receiver._name
#     #             sub_command = sub_command.__dict__
#     #     else:
#     #         # print(command.__dict__)
#     #         command._receiver = command._receiver._name
#     command_params = []
#     for command in command_list:
#         # if isinstance(command, CompositeCommand):
#         # print(type(command).__name__)
#         temp_dict_command_params = {"command": type(command).__name__}
#         temp_dict_command_params.update({"params": command.get_init_args()})
#         command_params.append(temp_dict_command_params)
#         # print(command._params)
#         # else:
#         #     command_params.append(command._params)
#     # print(command_params)
#     # print(com.get_command_names())
#     for index, command in enumerate(command_params):
#         # print(command)
#         children.append(
#             dbc.AccordionItem(
#                 dcc.Markdown(
#                     children=[
#                         "**Command Object:**",
#                         "```json",
#                         json.dumps(command, indent=4, cls=util.Encoder),
#                         # str(command.__dict__),
#                         "```",
#                     ],
#                 ),
#                 # str(command.__dict__),
#                 title=command["command"],
#                 item_id=command["command"] + str(index),
#             )
#         )
#     return children


# @app.callback(
#     Output("table-container3", "children"),
#     [Input("refresh-button3", "n_clicks")],
#     [State("table-container3", "children")],
# )
# def update_table3(n_clicks, table):
#     table_data3 = data_list3
#     table = dash_table.DataTable(
#         data=table_data3,
#         columns=[{"name": "Name", "id": "Name"}, {"name": "Value", "id": "Value"}],
#     )
#     return table



# modules for device as of commit 4123ed0

from typing import List

from devices.festo_solenoid_valve import FestoSolenoidValve
from .command import Command, CommandResult


class FestoParentCommand(Command):
    """Parent class for all Festo Solenoid Valve commands."""

    receiver_cls = FestoSolenoidValve

    def __init__(self, receiver: FestoSolenoidValve, **kwargs):
        super().__init__(receiver, **kwargs)


class FestoInitialize(FestoParentCommand):
    """Initialize the solenoid valve"""

    def __init__(self, receiver: FestoSolenoidValve, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())


class FestoDeinitialize(FestoParentCommand):
    """Deinitialize the solenoid valve"""

    def __init__(self, receiver: FestoSolenoidValve, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize())


class FestoValveOpen(FestoParentCommand):
    """Open the solenoid valve and keep open"""

    def __init__(self, receiver: FestoSolenoidValve, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.valve_open())


class FestoValveClosed(FestoParentCommand):
    """Close the solenoid valve and keep closed"""

    def __init__(self, receiver: FestoSolenoidValve, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.valve_closed())


class FestoOpenTimed(FestoParentCommand):
    """Open the valve for a set time then close"""

    def __init__(self, receiver: FestoSolenoidValve, time: int, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params["time"] = time

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.open_timed(self._params["time"]))
